"""
    The 'unit' must be ascii only.
"""
from collections import OrderedDict

import pytest
from pytest import approx

from barril import units
from barril.units import ChangeScalars
from barril.units import InvalidOperationError
from barril.units import InvalidUnitError
from barril.units import ObtainQuantity
from barril.units import Quantity
from barril.units import Scalar


def testScalarInterface(unit_database_well_length) -> None:
    s = Scalar("well-length")
    assert 0 == s.value
    assert "m" == s.unit


def testQuantity(unit_database_well_length) -> None:
    Quantity("well-length", "m")
    with pytest.raises(units.InvalidQuantityTypeError):
        Quantity("foo", "m")


def testQuantityEq(unit_database_well_length) -> None:
    q0 = ObtainQuantity("m", "well-length")
    q1 = ObtainQuantity("m", "well-length")
    assert q0 == q1


def testReadOnlyQuantityCopy(unit_database_well_length) -> None:
    q0 = ObtainQuantity("m", "well-length")
    q1 = q0.Copy()
    assert q0 == q1


def testScalar(unit_database_well_length) -> None:
    q = ObtainQuantity("m", "well-length")
    s = Scalar.CreateWithQuantity(q, 10)

    assert 10 == s.value
    assert approx(abs(393.700787402 - s.GetValue(unit="in")), 7) == 0

    s = s.CreateCopy(value=0.254, unit="in")
    assert 0.254 == s.value
    assert "in" == s.unit
    assert "well-length" == s.category

    with pytest.raises(AttributeError):
        setattr(s, "value", 10)


def GetQuantity(unit="cm"):
    return units.ObtainQuantity(unit, "well-length")


def testBadUnit(unit_database_well_length) -> None:
    x = units.Scalar.CreateWithQuantity(GetQuantity(), 500)
    with pytest.raises(units.InvalidUnitError):
        x.CreateCopy(unit="invalidUnit")


def testOperations(unit_database_well_length) -> None:
    x = units.Scalar.CreateWithQuantity(GetQuantity(), 500)
    assert x.value == 500
    assert x.unit == "cm"

    assert x.GetValue("m") == 5
    assert x.unit == "cm"
    assert x.GetValue("cm") == 500
    assert x.unit == "cm"
    assert x.GetValue("km") == 0.005


def unit(value, unit="m"):
    return units.Scalar.CreateWithQuantity(GetQuantity(unit), value)


def testEq(unit_database_well_length) -> None:
    u1 = units.Scalar.CreateWithQuantity(GetQuantity(), 1.0)
    u2 = units.Scalar.CreateWithQuantity(GetQuantity(), 1.0)
    assert u1 == u2
    assert u1.unit == u2.unit
    assert "cm" == u2.unit


def testAddUnit(unit_database_empty) -> None:
    unit_database = unit_database_empty
    unit_database.AddUnitBase("length", "metre", "m")
    RATIO = 0.0254
    unit_database.AddUnit(
        "length", "inch", "in", lambda x: 0.0254 * RATIO, lambda y: y / RATIO, default_category=None
    )
    unit_database.AddCategory("length", "length", override=True, valid_units=["m", "in"])
    unit_database.AddCategory("depth", "length", override=True, valid_units=["m", "in"])

    a = units.Scalar("length", value=1.0, unit="m")
    b = units.Scalar("length", value=2.0, unit="m")
    c = units.Scalar("length", value=1.0, unit="in")
    d = units.Scalar("depth", value=1.0, unit="m")
    e = None
    f = units.Scalar("length", value=1.0, unit="m")

    assert a != b
    assert a != c
    assert a != e
    assert a != d

    assert a == f


def testValidUnits(unit_database_empty) -> None:
    # let'scalar clear up the unit manager
    unit_database = unit_database_empty

    unit_database.AddUnitBase("length", "meters", "m")
    unit_database.AddUnit("length", "centimeters", "cm", "%f * 100.0", "%f / 100.0")
    unit_database.AddUnit("length", "milimeters", "mm", "%f * 1000.0", "%f / 1000.0")
    unit_database.AddUnit("length", "kilometers", "km", "%f / 1000.0", "%f * 1000.0")
    unit_database.AddUnit("length", "miles", "mi", "%f / 1609.347", "%f * 1609.347")
    unit_database.AddUnit("length", "inches", "in", "%f / 0.0254", "%f * 0.0254")

    valid_units = ["m", "cm", "mm"]
    unit_database.AddCategory("length", "length")
    unit_database.AddCategory("well-length", "length", valid_units=valid_units)
    unit_database.AddCategory("well-diameter", "length")

    Quantity("well-length", "km")

    scalar = units.Scalar("well-length", 1, "m")
    assert valid_units == scalar.GetValidUnits()

    scalar = units.Scalar("well-diameter", 1, "m")
    assert ["m", "cm", "mm", "km", "mi", "in"] == scalar.GetValidUnits()

    # Creating a scalar in a unit that isn't valid shouldn't raise an error and the unit
    # will be added to valid units
    scalar = units.Scalar("well-length", 1, "mi")
    assert scalar.GetValidUnits() == ["m", "cm", "mm", "mi"]


def testScalarCopyAndRepresentation(unit_database_empty) -> None:
    unit_database = unit_database_empty

    unit_database.AddUnitBase("length", "meters", "m")
    unit_database.AddUnit("length", "milimeters", "mm", "x * 1000.0", "x / 1000.0")
    unit_database.AddUnit("length", "centimeters", "cm", "x * 100.0", "x / 100.0")
    unit_database.AddCategory("well-diameter", "length")
    s = Scalar("well-diameter", 10, "m")

    assert "Scalar(10.0, 'm', 'well-diameter')" == repr(s)
    assert "10.0" == str(s.value)
    assert "m" == s.unit
    assert "well-diameter" == s.GetCategory()
    assert "length" == s.GetQuantityType()

    s = s.CreateCopy(unit="cm")
    assert "Scalar(1000.0, 'cm', 'well-diameter')" == repr(s)
    assert "1000 [cm]" == str(s)


def testSort(unit_database_empty) -> None:
    db = unit_database_empty

    db.AddUnitBase("length", "meters", "m")
    db.AddUnit("length", "centimeters", "cm", "x * 100.0", "x / 100.0")
    db.AddCategory("length", "length")

    s1 = Scalar("length", 10.0, "cm")
    s2 = Scalar("length", 1.0, "m")
    s3 = Scalar("length", 500.0, "cm")
    s4 = Scalar("length", 10.0, "m")

    x = [s2, s1, s4, s3]
    x.sort()
    assert x == [s1, s2, s3, s4]


def testCategoryParameters(unit_database_empty) -> None:
    db = unit_database_empty

    db.AddUnitBase("length", "meters", "m")
    db.AddUnit("length", "centimeters", "cm", "x * 100.0", "x / 100.0")

    db.AddCategory(
        "my length",
        "length",
        ["cm", "m"],
        default_unit="m",
        default_value=18.5,
        min_value=-30.0,
        max_value=100.0,
        is_min_exclusive=True,
        is_max_exclusive=False,
    )

    s1 = Scalar("my length")
    assert s1.value == 18.5
    assert s1.unit == "m"

    # test maximum boundary
    s2 = Scalar("my length", 100.0, "m")
    assert s2.value == 100.0
    assert s2.unit == "m"

    s3 = Scalar("my length", 500.0, "cm")
    assert s3.value == 500.0
    assert s3.unit == "cm"

    # higher than maximum value
    another = Scalar("my length", 120.0, "m")
    assert not another.IsValid()
    with pytest.raises(ValueError):
        another.CheckValidity()

    another = Scalar("my length", -30.0, "m")
    assert not another.IsValid()

    s4 = Scalar("my length", unit="cm")
    assert s4.value == 1850.0
    assert s4.unit == "cm"

    another = s4.CreateCopy(12500.00, "cm")
    assert not another.IsValid()


def testInvalidUnit(unit_database_empty) -> None:
    unit_database = unit_database_empty

    unit_database.AddUnitBase("length", "meters", "m")
    unit_database.AddUnit("length", "milimeters", "mm", "x * 1000.0", "x / 1000.0")
    unit_database.AddUnit("length", "centimeters", "cm", "x * 100.0", "x / 100.0")
    unit_database.AddCategory("well-diameter", "length")
    with pytest.raises(InvalidUnitError):
        Scalar("well-diameter", 10, "days")


def testDerivedUnits(unit_database_empty) -> None:
    unit_database = unit_database_empty

    unit_database.AddUnitBase("length", "meters", "m")
    unit_database.AddUnit("length", "milimeters", "mm", "x * 1000.0", "x / 1000.0")
    unit_database.AddUnit("length", "centimeters", "cm", "x * 100.0", "x / 100.0")
    unit_database.AddCategory("well-diameter", "length")
    unit_database.AddUnitBase("time", "seconds", "s")
    unit_database.AddCategory(category="Time", quantity_type="time")

    s1 = Scalar("well-diameter", 10, "m")
    s2 = Scalar("well-diameter", 10, "cm")
    s3 = Scalar("Time", 10, "s")

    assert Scalar("well-diameter", 10.10, "m") == s1 + s2
    assert Scalar("well-diameter", 9.90, "m") == s1 - s2

    quantity = Quantity.CreateDerived(OrderedDict())
    assert Scalar.CreateWithQuantity(quantity, 100) == s1 / s2
    assert Scalar("well-diameter", 9.90, "m") == s1 - s2
    with pytest.raises(InvalidOperationError):
        s1.__sub__(s3)


def testCreationWithDerivedQuantity(unit_database_len_time) -> None:
    unit_database = unit_database_len_time

    m = Quantity.CreateDerived(OrderedDict([("Table size", ["m", 1])]))
    km_city = Quantity.CreateDerived(OrderedDict([("City size", ["km", 1])]))

    quantity, value = unit_database.Multiply(m, km_city, 1, 0.01)
    calculated1 = Scalar.CreateWithQuantity(quantity, value)
    assert str(calculated1)

    s1 = Scalar.CreateWithQuantity(m, 1)
    s2 = Scalar.CreateWithQuantity(km_city, 0.01)
    assert calculated1 == s1 * s2


def testNumberInteractions(unit_database_len_time) -> None:
    scalar = Scalar("Table size", 1, "m")
    scalar2 = Scalar.CreateWithQuantity(Quantity.CreateDerived(OrderedDict()), 0)

    assert scalar == scalar + scalar2

    assert scalar == scalar + 0
    assert scalar == 0 + scalar

    assert 9 == (10 - scalar).value
    assert -9 == (scalar - 10).value

    assert 10 == (10 * scalar).value
    assert 10 == (scalar * 10).value

    assert 10 == (10.0 / scalar).value
    assert 1 / 10.0 == (scalar / 10.0).value


def testDivision(unit_database_len_time) -> None:
    unit_database = unit_database_len_time

    m = Quantity.CreateDerived(OrderedDict([("Table size", ["m", 1])]))
    km_city = Quantity.CreateDerived(OrderedDict([("City size", ["km", 1])]))
    quantity, value = unit_database.Divide(m, km_city, 1, 0.01)
    calculated1 = Scalar.CreateWithQuantity(quantity, value)
    s1 = Scalar.CreateWithQuantity(m, 1)
    s2 = Scalar.CreateWithQuantity(km_city, 0.01)
    assert calculated1 == s1 / s2


def testFloorDivision() -> None:
    a = Scalar(3.5, "m")
    b = Scalar(100.0, "cm")
    assert (a // b).GetValueAndUnit() == (3.0, "")
    assert (350 // b).GetValueAndUnit() == (3.0, "1/cm")
    assert (a // 1.0).GetValueAndUnit() == (3.0, "m")


def testNumberOverScalar() -> None:
    a = Scalar(2.0, "m")
    b = Scalar(3.0, "m")
    assert (1.0 / a).GetValueAndUnit() == (0.5, "1/m")
    assert (3.0 / a).GetValueAndUnit() == (1.5, "1/m")
    assert b / a == b * 1 / a == b * (1.0 / a)


def testPow(unit_database_len_time) -> None:
    s = Scalar(2, "s", "Time")
    spow = s**3
    assert spow.unit == "s3"
    assert spow.value == 8


def testFormatedUnitsOnScalar(unit_database_empty) -> None:
    """
    Allow present units on get formatted Scalar
    """
    unit_database = unit_database_empty
    unit_database.AddUnitBase("length", "meters", "m")
    unit_database.AddCategory(category="length", quantity_type="length")

    class MyScalar(units.Scalar):
        pass

    scalar = MyScalar("length", 1.18, "m")
    assert scalar.GetFormatted() == "1.18 [m]"

    MyScalar.SetFormattedValueFormat("%0.4f")
    MyScalar.SetFormattedSuffixFormat(" (%s)")

    # Tests the GET methods
    assert MyScalar.GetFormattedValueFormat() == "%0.4f"
    assert MyScalar.GetFormattedSuffixFormat() == " (%s)"

    # Tests the Formatted method
    assert scalar.GetFormatted() == "1.1800 (m)"

    # Check for invalid format strings
    with pytest.raises(TypeError):
        MyScalar.SetFormattedValueFormat("%d [%s]")
    with pytest.raises(TypeError):
        MyScalar.SetFormattedSuffixFormat("%s %s")
    with pytest.raises(TypeError):
        MyScalar.SetFormattedSuffixFormat("%f")

    # When invalid format are set, the formats are not changed
    assert MyScalar.GetFormattedValueFormat() == "%0.4f"
    assert MyScalar.GetFormattedSuffixFormat() == " (%s)"

    # Make sure we didn't affect the Scalar class, only the local MyScalar
    MyScalar.SetFormattedValueFormat("<<< %f")
    MyScalar.SetFormattedSuffixFormat(" [%s] >>>")
    scalar2 = Scalar("length", 1.18, "m")
    assert scalar2.GetFormatted() == "1.18 [m]"


def testEmptyScalar() -> None:
    """
    ScalarMultiData exception when some of its scalars don't have a quantity_type
    """
    # An empty scalar doesn't have a category defined
    scalar_1 = Scalar.CreateEmptyScalar(20.0)

    # When try to retrieve scalar value or unit a  exception was being raised
    assert scalar_1.GetValue("m") == 20.0
    assert scalar_1.GetUnit() == ""


def testCopyProperties(unit_database_well_length) -> None:
    """
    Test if the mehod SetValue is not called when copying the Scalar's properties.
    """
    category = "well-length"
    unit = "m"
    value = 10.0

    scalar_source = Scalar(category, value, unit)
    scalar_dest = scalar_source.CreateCopyInstance()

    assert scalar_dest.GetCategory() == category
    assert scalar_dest.GetUnit() == unit
    assert scalar_dest.GetValue() == value


def testCopyPropertiesAndValidation(unit_database_well_length) -> None:
    category = "well-length-with-min-and-max"  # the minimum value is zero
    unit = "m"
    value = -1

    # Not raises exception because validation is False on the copy operation
    scalar_source = Scalar(category, value, unit)
    assert not scalar_source.IsValid()

    # Raises exception because validation is True and the value -1 is invalid accordingly to
    # the category 'well-length-with-min-and-max'
    another = scalar_source.CreateCopy()
    assert not another.IsValid()


def testDefaultValue(unit_database_len_pressure) -> None:
    """
    Scalar constructor considers the minimum and maximum values
    when default_value is not defined
    """
    db = unit_database_len_pressure

    db.AddCategory(category="my length", quantity_type="length", min_value=100.0, max_value=200.0)

    # if the default value is not defined, the scalar should not try to set the initial/first
    # value of the new instance to 0.0, but it should assume the minimum allowed value.
    # should not raise ValueError
    length = Scalar("my length")
    assert length.GetValue() == 100.0

    length = Scalar(ObtainQuantity("m", "my length"))
    assert length.GetValue() == 100.0

    length = Scalar(ObtainQuantity("m"))
    assert length.GetValue() == 0.0

    # invalid default value (< min)
    with pytest.raises(AssertionError):
        db.AddCategory(
            category="my pressure",
            quantity_type="pressure",
            default_value=50.0,
            min_value=100.0,
            max_value=200.0,
        )

    # invalid default value (> max)
    with pytest.raises(AssertionError):
        db.AddCategory(
            category="my pressure",
            quantity_type="pressure",
            default_value=300.0,
            min_value=100.0,
            max_value=200.0,
        )

    # invalid default value (<= min)
    with pytest.raises(AssertionError):
        db.AddCategory(
            category="my pressure",
            quantity_type="pressure",
            default_value=100.0,
            min_value=100.0,
            max_value=200.0,
            is_min_exclusive=True,
        )

    # invalid default value (>= min)
    with pytest.raises(AssertionError):
        db.AddCategory(
            category="my pressure",
            quantity_type="pressure",
            default_value=200.0,
            min_value=100.0,
            max_value=200.0,
            is_max_exclusive=True,
        )

    db.AddCategory(
        category="my pressure",
        quantity_type="pressure",
        min_value=100.0,
        max_value=200.0,
        default_value=150.0,
        is_min_exclusive=True,
        is_max_exclusive=False,
    )

    pressure = Scalar("my pressure")
    assert pressure.GetValue() == 150.0

    # default_value not informed. checking if the interval limits are respected
    # when is_min_exclusive or is_max_exclusive have been informed, so the default_value
    # must be
    with pytest.raises(RuntimeError):
        db.AddCategory(
            category="my pressure 2",
            quantity_type="pressure",
            min_value=100.0,
            max_value=200.0,
            is_max_exclusive=True,
        )


def testScalarInvalidValue(unit_database_len_time) -> None:
    db = unit_database_len_time

    db.AddCategory("another-length", "length", min_value=0, max_value=15)

    scalar = Scalar("another-length", value=15, unit="m")
    assert scalar.IsValid()

    scalar = Scalar("another-length", value=-5, unit="m")
    assert not scalar.IsValid()

    # By default the validation will be performed. 15 is a valid value.
    scalar = Scalar("another-length", value=15, unit="m")
    assert scalar.IsValid()

    # Even invalid ,the scalar returns the value, unit and a formatted text.
    another = Scalar("another-length", value=3000, unit="m")
    assert not another.IsValid()
    assert another.GetValue("m") == 3000
    assert another.GetUnit() == "m"
    assert another.GetFormatted() == "3000 [m]"

    # By default the validation will be performed, and in this cases will raise ValueError.
    another_2 = Scalar("another-length", unit="m", value=5000)
    assert not another_2.IsValid()

    # Performing copy between invalid scalars. The validation is not performed on copy.
    copied = another.CreateCopy(unit="cm")
    assert not copied.IsValid()
    assert copied.GetValue("m") == 3000
    assert copied.GetUnit() == "cm"
    assert copied.GetFormatted() == "300000 [cm]"


def testScalarValidUnits(unit_database_empty) -> None:
    db = unit_database_empty
    db.AddUnit("length", "milimeters", "mm", "%f * 1000.0", "%f / 1000.0")
    db.AddUnitBase("length", "meters", "m")
    db.AddUnit("length", "centimeters", "cm", "%f * 100.0", "%f / 100.0")
    db.AddUnit("length", "kilometers", "km", "%f / 1000.0", "%f * 1000.0")
    db.AddCategory("len1", "length", valid_units=["mm", "cm"])
    db.AddCategory("len2", "length", valid_units=["m", "km"])

    s1 = Scalar("len1", 1, "mm")
    assert s1 is not None

    s2 = Scalar("len2", 1, "mm")
    assert s2 is not None


def testScalarRepresentation(unit_database_posc) -> None:
    s1 = Scalar(0.5, "kg/kg", "dimensionless")
    s2 = Scalar(0.125, "g/g", "mass concentration")

    assert s1.GetCategory() == "dimensionless"
    assert s2.GetCategory() == "mass concentration"

    assert repr(s1) == "Scalar(0.5, 'kg/kg', 'dimensionless')"
    assert repr(s2) == "Scalar(0.125, 'g/g', 'mass concentration')"


def testScalarCreationModes() -> None:
    base = Scalar(10, "m")

    assert Scalar("length", 10, "m") == base
    assert Scalar((10.0, "m")) == base

    with pytest.raises(AssertionError):
        Scalar("length", 1.0)  # missing unit


def testScalarPickle(unit_database_posc) -> None:
    import pickle

    simple_scalar = Scalar("length", 10, "m")
    simple_scalar2 = pickle.loads(pickle.dumps(simple_scalar))
    assert simple_scalar == simple_scalar2

    complex_scalar = Scalar("length", 10, "m") * Scalar("time", 5, "s")
    complex_scalar2 = pickle.loads(pickle.dumps(complex_scalar))
    assert complex_scalar == complex_scalar2


def testScalarHashEq() -> None:
    scalar1 = Scalar("length", 10, "m")
    scalar2 = Scalar("length", 10, "m")
    scalar3 = Scalar("length", 10, "cm")
    scalar4 = Scalar("length", scalar2.GetValue("cm"), "cm")

    assert scalar1 == scalar2
    assert hash(scalar1) == hash(scalar2)
    assert hash(scalar1) == hash(scalar1)
    assert scalar2 != scalar3
    assert hash(scalar2) != hash(scalar3)
    assert scalar3 != scalar4
    assert hash(scalar4) != hash(scalar3)


def testChangeScalars() -> None:
    class Fluid:
        density = Scalar(0, "lbm/galUS")
        concentration = Scalar(0, "%")

    fluid = Fluid()
    assert fluid.density.GetValue("lbm/galUS") == 0
    assert fluid.concentration.GetValue("%") == 0

    ChangeScalars(fluid, density=(10, "lbm/galUS"), concentration=(1.0, "%"))

    assert fluid.density.GetValue("lbm/galUS") == 10
    assert fluid.concentration.GetValue("%") == 1.0



from docutils import nodes
from docutils.statemachine import ViewList
from sphinx.util import nested_parse_with_titles
from sphinx.util.docutils import SphinxDirective

from barril.units import UnitDatabase


class ListAllUnits(SphinxDirective):
    def run(self):
        rst = ViewList()
        unit_database = UnitDatabase.GetSingleton()
        source = "fakefile.rst"

        # Create the rst content
        for category in sorted(unit_database.IterCategories(), key=str.casefold):
            title = unit_database.GetCategoryInfo(category).caption.title()
            rst.append(f".. rubric:: {title}", source)
            rst.append(f"", source)

            for unit in sorted(
                unit_database.GetCategoryInfo(category).valid_units_set, key=str.casefold
            ):
                name = unit_database.unit_to_unit_info[unit].name
                rst.append(f'- ``"{unit}"`` ({name})', source)

            rst.append(f" ", source)

        # Create a node.
        node = nodes.section()
        node.document = self.state.document

        # Parse the rst.
        nested_parse_with_titles(self.state, rst, node)

        # And return the result.
        return node.children


def setup(app):
    app.add_directive("list_all_units", ListAllUnits)
    return {"version": "0.1", "parallel_read_safe": True, "parallel_write_safe": True}



import pytest
from pytest import approx

from barril import units


@pytest.fixture
def db():
    db = units.UnitDatabase.GetSingleton()
    yield db


def testTransmissibility(db) -> None:
    converted = db.Convert("transmissibility", "cp.m3/day/bar", "cp.bbl/day/psi", 1.0)
    assert approx(converted) == 0.433667315


def testIndex(db) -> None:
    category = db.GetCategoryInfo("index")
    db.CheckQuantityTypeUnit(category.quantity_type, "<ind>")


def testAdsorptionRate(db) -> None:
    db.CheckQuantityTypeUnit("adsorption rate", "mg/kg/d")
    assert approx(db.Convert("adsorption rate", "mg/kg/d", "kg/kg/d", 1.0)) == 1.0e-6


def testSolubilityProduct(db) -> None:
    db.CheckQuantityTypeUnit("solubility product", "(mol/m3)^2")
    assert approx(db.Convert("solubility product", "(mol/L)^2", "(mol/m3)^2", 1.0)) == 1.0e6


def testMassConsumptionEfficiency(db) -> None:
    db.CheckQuantityTypeUnit("mass consumption efficiency", "mg/l/mg/l")
    db.CheckQuantityTypeUnit("mass consumption efficiency", "kg/m3/kg/m3")


def testDensityGeneration(db) -> None:
    db.CheckQuantityTypeUnit("density generation", "mg/l/d")
    assert approx(db.Convert("density generation", "mg/l/d", "kg/m3/d", 1.0)) == 0.001


def testMolarDensity(db) -> None:
    db.CheckQuantityTypeUnit("amount of a substance", "gmole/m3")
    db.CheckQuantityTypeUnit("concentration of B", "gmole/m3")


def testPartsPerMillionByVolumeConcentration(db) -> None:
    assert (
        approx(
            db.Convert(
                "parts per million by volume per concentration", "ppmv/mg/l", "ppmv/kg/m3", 1.0
            )
        )
        == 1000
    )


def testKPaPerSecond(db) -> None:
    assert approx(db.Convert("dynamic viscosity", "Pa.s", "kPa.s", 1000.0)) == 1


def testPartsPerMillionVolume(db) -> None:
    quantity_type = db.GetCategoryInfo("relative proportion").quantity_type
    db.CheckQuantityTypeUnit(quantity_type, "ppmv")


def testPerTimeSquared(db) -> None:
    quantity_type = db.GetCategoryInfo("per time squared").quantity_type
    db.CheckQuantityTypeUnit(quantity_type, "1/d^2")


def testNoUnit(db) -> None:
    db.CheckQuantityTypeUnit("dimensionless", "-")


def testFluidConsistencyQuantityAndConversions(db) -> None:
    # Checking the quantity type.
    db.CheckQuantityType("fluid consistency")

    # Checking the units on quantity type.
    db.CheckQuantityTypeUnit("fluid consistency", "Pa.s^n")
    db.CheckQuantityTypeUnit("fluid consistency", "lbf.s^n/ft2")
    db.CheckQuantityTypeUnit("fluid consistency", "eq.cp")

    # Let's do some conversions.
    def DoConversion(from_unit, to_unit, value):
        return db.Convert("fluid consistency", from_unit, to_unit, value)

    assert approx(DoConversion("Pa.s^n", "lbf.s^n/ft2", 1.0)) == 0.02088543
    assert approx(DoConversion("Pa.s^n", "lbf.s^n/ft2", 5.0)) == 0.10442717
    assert approx(DoConversion("Pa.s^n", "lbf.s^n/ft2", 5.5)) == 0.11486989
    assert approx(DoConversion("Pa.s^n", "lbf.s^n/ft2", 250.0)) == 5.221358447

    assert approx(DoConversion("Pa.s^n", "lbf.s^n/100ft2", 1.0)) == 2.08854338
    assert approx(DoConversion("Pa.s^n", "lbf.s^n/100ft2", 5.0)) == 10.44271689
    assert approx(DoConversion("Pa.s^n", "lbf.s^n/100ft2", 5.5)) == 11.48698858
    assert approx(DoConversion("Pa.s^n", "lbf.s^n/100ft2", 250.0)) == 522.1358447

    assert approx(DoConversion("Pa.s^n", "eq.cp", 1.0)) == 994.5238095
    assert approx(DoConversion("Pa.s^n", "eq.cp", 5.0)) == 4972.6190475
    assert approx(DoConversion("Pa.s^n", "eq.cp", 5.5)) == 5469.8809522
    assert approx(DoConversion("Pa.s^n", "eq.cp", 250.0)) == 248630.952375

    assert approx(DoConversion("lbf.s^n/ft2", "Pa.s^n", 1.0)) == 47.88026
    assert approx(DoConversion("lbf.s^n/ft2", "Pa.s^n", 5.0)) == 239.4013
    assert approx(DoConversion("lbf.s^n/ft2", "Pa.s^n", 5.5)) == 263.34143
    assert approx(DoConversion("lbf.s^n/ft2", "Pa.s^n", 250.0)) == 11970.065

    assert approx(DoConversion("lbf.s^n/ft2", "lbf.s^n/100ft2", 1.0)) == 100
    assert approx(DoConversion("lbf.s^n/ft2", "lbf.s^n/100ft2", 5.0)) == 500
    assert approx(DoConversion("lbf.s^n/ft2", "lbf.s^n/100ft2", 5.5)) == 550
    assert approx(DoConversion("lbf.s^n/ft2", "lbf.s^n/100ft2", 250.0)) == 25000

    assert approx(DoConversion("lbf.s^n/ft2", "eq.cp", 1.0)) == 47618.058575050469

    assert approx(DoConversion("eq.cp", "Pa.s^n", 1.0), abs=1e7) == 0.0010055
    assert approx(DoConversion("eq.cp", "Pa.s^n", 5.0), abs=1e7) == 0.0050275
    assert approx(DoConversion("eq.cp", "Pa.s^n", 5.5), abs=1e7) == 0.0055303
    assert approx(DoConversion("eq.cp", "Pa.s^n", 250.0)) == 0.2513766

    assert approx(DoConversion("eq.cp", "lbf.s^n/ft2", 1.0), abs=1e6) == 0.000021
    assert approx(DoConversion("eq.cp", "lbf.s^n/ft2", 5.0), abs=1e6) == 0.000105
    assert approx(DoConversion("eq.cp", "lbf.s^n/ft2", 5.5), abs=1e6) == 0.0001155
    assert approx(DoConversion("eq.cp", "lbf.s^n/ft2", 250.0)) == 0.00525011

    assert approx(DoConversion("eq.cp", "lbf.s^n/100ft2", 1.0), abs=1e7) == 0.0021
    assert approx(DoConversion("eq.cp", "lbf.s^n/100ft2", 5.0)) == 0.01050021
    assert approx(DoConversion("eq.cp", "lbf.s^n/100ft2", 5.5)) == 0.011550239
    assert approx(DoConversion("eq.cp", "lbf.s^n/100ft2", 250.0)) == 0.5250109


def testBblPerMeters(db) -> None:
    assert approx(db.Convert("volume per length", "bbl/ft", "bbl/m", 1)) == 1 / 0.3048


def testBblPerSecond(db) -> None:
    assert approx(db.Convert("volume flow rate", "bbl/hr", "bbl/s", 1)) == 1 / 3600


def testConcentration(db) -> None:
    assert approx(db.Convert("concentration", "g/L", "mg/L", 1)) == 1000.0


def testVolumeFractionPerTemperature(db) -> None:
    expected_result = db.Convert("volumetric thermal expansion", "1/degC", "1/degF", 1)
    obtained_result = db.Convert(
        "volume fraction per temperature", "(m3/m3)/degC", "(m3/m3)/degF", 1
    )
    assert approx(obtained_result) == expected_result

    expected_units = ["(m3/m3)/K", "(m3/m3)/degC", "(m3/m3)/degF"]
    obtained_result = sorted(db.GetValidUnits("volume fraction per temperature"))
    assert obtained_result == expected_units


def testVolumePerWtpercent(db) -> None:
    ft3_wtpercent_to_m3_wtpercent = db.Convert(
        "volume per wtpercent", "ft3/wtpercent", "m3/wtpercent", 25 / 1
    )
    m3_to_wtpercent_to_ft3_to_wtpercent = db.Convert(
        "volume per wtpercent", "m3/wtpercent", "ft3/wtpercent", 0.70792125
    )

    assert approx(ft3_wtpercent_to_m3_wtpercent) == 0.70792125
    assert approx(m3_to_wtpercent_to_ft3_to_wtpercent) == 25


def testMassPerMol(db) -> None:
    gpermol_to_kgpermol = db.Convert("mass per mol", "g/mol", "kg/mol", 1000 / 1)
    kgpermol_to_gpermol = db.Convert("mass per mol", "kg/mol", "g/mol", 1)

    assert gpermol_to_kgpermol == 1
    assert kgpermol_to_gpermol == 1000
    assert db.Convert("mass per mol", "g/mol", "lb/lbmole", 1) == 1


def testInjectivityFactor(db) -> None:
    quantity_type = db.GetCategoryInfo("injectivity factor").quantity_type

    db.CheckQuantityType(quantity_type)
    db.CheckQuantityTypeUnit(quantity_type, "m3/s.Pa")

    expected_units = ["m3/s.Pa", "bbl/min.psi", "(m3/d)/(kgf/cm2)"]

    assert db.GetValidUnits("injectivity factor") == expected_units

    converted_value = db.Convert("injectivity factor", "bbl/min.psi", "m3/s.Pa", 1)
    # (from bbl/min to m3/s) / (from psi to Pa)
    assert approx(converted_value) == 0.002649 / 6894.757

    converted_value = db.Convert("injectivity factor", "(m3/d)/(kgf/cm2)", "m3/s.Pa", 1)
    assert approx(converted_value) == 1.1802270983e-10


def testTemperaturePerLengthdegCP30m() -> None:
    u1 = units.Scalar("temperature per length", 1.0, "K/m")
    u2 = units.Scalar("temperature per length", 30.0, "degC/30m")

    assert approx(u1.GetValue("degC/30m")) == 30.0
    assert approx(u2.GetValue("K/m")) == 1.0


def testMassPerTimePerArea(db) -> None:
    secs_to_days = 60 * 60 * 24
    assert approx(db.Convert("mass per time per area", "kg/m2.s", "kg/m2.d", 1)) == 1 * secs_to_days


def testAccelerationLinearMetersPerMin2(db) -> None:
    assert approx(db.Convert("acceleration linear", "m/s2", "m/min2", 2)) == 7200
    assert approx(db.Convert("acceleration linear", "m/min2", "m/s2", 10800)) == 3


def testAccelerationLinearFeetPerMin2(db) -> None:
    assert approx(db.Convert("acceleration linear", "ft/s2", "ft/min2", 2)) == 7200
    assert approx(db.Convert("acceleration linear", "ft/min2", "ft/s2", 10800)) == 3


def testMassConcentration(db) -> None:
    assert approx(db.Convert("mass concentration", "Euc", "g/g", 2)) == 2
    assert approx(db.Convert("mass concentration", "Euc", "g/100g", 2)) == 200
    assert approx(db.Convert("mass concentration", "Euc", "lbm/lbm", 2)) == 2


def testSpecificVolume(db) -> None:
    assert approx(db.Convert("specific volume", "m3/kg", "l/mg", 1.0)) == 0.001


def testViscosityPerPressure(db) -> None:
    assert approx(db.Convert("viscosity per pressure", "cP/kPa", "cP/psi", 10.0)) == 68.9475729

    assert approx(db.Convert("viscosity per pressure", "cP/kPa", "cP/(kgf/cm2)", 10.0)) == 980.665

    assert (
        approx(db.Convert("viscosity per pressure", "cP/(kgf/cm2)", "cP/psi", 0.102), abs=1e7)
        == 0.0071713
    )

    # maybe it's from this wrong conversion where IMEX limit of 0.102 comes from:
    assert approx(db.Convert("viscosity per pressure", "cP/(kgf/cm2)", "cP/kPa", 10.0)) == 0.1019716


def testForcePerLength(db) -> None:
    assert approx(db.Convert("force per length", "N/m", "kgf/m", 1.0)) == 0.101971621


def testMiligramPerGram(db) -> None:
    assert approx(db.Convert("mass concentration", "mg/g", "g/g", 2.0)) == 0.002


def testGramsPerBarrel(db) -> None:
    db.CheckQuantityTypeUnit("concentration", "g/bbl")


def testSiemens(db) -> None:
    assert approx(db.Convert("conductivity", "S/m", "uS/m", 1.0)) == 1000000.0


def testStrokeFrequency(db) -> None:
    db.CheckQuantityTypeUnit("stroke frequency", "spm")
    assert approx(db.Convert("stroke frequency", "spm", "sps", 150.0)) == 2.5


def testPowerPerWeight(db) -> None:
    db.CheckQuantityTypeUnit("power per mass", "W/kg")
    db.CheckQuantityTypeUnit("power per weight", "W/kg")

    assert approx(db.Convert("power per weight", "W/kg", "kW/kg", 1.0)) == 1.0 / 1000.0
    assert approx(db.Convert("power per weight", "kW/kg", "W/kg", 1.0)) == 1000.0


def testSelfInductance(db) -> None:
    assert approx(db.Convert("self inductance per length", "H/m", "H/km", 1.0), rel=0.01) == 1000

    assert (
        approx(db.Convert("self inductance per length", "H/m", "mH/km", 1.0), rel=0.01) == 1000000
    )

    assert (
        approx(db.Convert("self inductance per length", "H/m", "uH/km", 1.0), rel=0.01)
        == 1000000000
    )

    assert (
        approx(db.Convert("self inductance per length", "H/m", "nH/km", 1.0), rel=0.01)
        == 1000000000000
    )

    assert approx(db.Convert("self inductance per length", "H/m", "mH/m", 1.0), rel=0.01) == 1000

    assert approx(db.Convert("self inductance per length", "H/m", "uH/m", 1.0), rel=0.01) == 1000000

    assert (
        approx(db.Convert("self inductance per length", "H/m", "nH/m", 1.0), rel=0.01) == 1000000000
    )


def testConcentrationPerSquareTime(db) -> None:
    db.CheckQuantityTypeUnit("concentration per square time", "mg/l/d2")
    db.CheckQuantityTypeUnit("concentration per square time", "kg/m3/d2")

    assert (
        approx(db.Convert("concentration per square time", "mg/l/d2", "kg/m3/d2", 1.0))
        == 1.0 / 1000.0
    )
    assert approx(db.Convert("concentration per square time", "kg/m3/d2", "mg/l/d2", 1.0)) == 1000.0


def testVolumeInCubicMicrometres(db) -> None:
    obtained = db.Convert("volume", "m3", "um3", 1.0)
    expected = 1e18
    assert approx(obtained, rel=1e-12) == expected


def testVolumeFlowRateInCubicMicrometresPerSecond(db) -> None:
    obtained = db.Convert("volume flow rate", "m3/s", "um3/s", 1.0)
    expected = 1e18
    assert approx(obtained, rel=1e-12) == expected


def testFlowCoefficient(db) -> None:
    obtained = db.Convert("flow coefficient", "(galUS/min)/(psi^0.5)", "(m3/s)/(Pa^0.5)", 1.0)
    expected = 7.59805421208337e-07
    assert approx(obtained, rel=1e-12) == expected

    obtained = db.Convert("flow coefficient", "(m3/h)/(bar^0.5)", "(m3/s)/(Pa^0.5)", 1.0)
    expected = 8.784104611578831e-07
    assert approx(obtained, rel=1e-12) == expected


def testHertzPerSecond() -> None:
    from barril.units import Scalar

    assert approx(Scalar(1, "rpm").GetValue("Hz")) == 1 / 60.0
    assert approx(Scalar(1, "Hz").GetValue("rpm")) == 60.0
    assert approx(Scalar(1, "Hz/s").GetValue("rpm/s")) == 60.0
    assert approx(Scalar(1, "rpm/s").GetValue("Hz/s")) == 1 / 60.0


def testFluidGasConcentration() -> None:
    from barril.units import Scalar

    assert approx(Scalar(1, "tgu").GetValue("ppm")) == 333.33
    assert approx(Scalar(1, "tgu").GetValue("%")) == 0.033333


def testPerMicrometre(db) -> None:
    per_micrometre = units.Scalar("per length", 1.0, "1/um")
    per_metre = per_micrometre.CreateCopy(unit="1/m")
    per_inch = per_micrometre.CreateCopy(unit="1/in")

    assert per_micrometre.GetValue("1/m") == 10**6
    assert approx(per_metre.GetValue()) == 10**6
    assert approx(per_inch.GetValue()) == 25400
    assert per_metre == units.Scalar("per length", 10**6, "1/m")
    assert per_inch == units.Scalar("per length", 25400.0, "1/in")



from typing import Any
from typing import overload
from typing import Tuple
from typing import Union

from oop_ext.interface import Interface
from oop_ext.interface import TypeCheckingSupport

from barril.units import Array
from barril.units._array import ValuesType

__all__ = ["ICurve"]


class ICurve(Interface, TypeCheckingSupport):
    """
    The curve is an element that has values and domain for those values.
    """

    def GetImage(self) -> Array:  # type:ignore[empty-body]
        """
        :returns:
            An IArray -- which is an IObjectWithQuantity with the image for this curve
        """

    def GetDomain(self) -> Array:  # type:ignore[empty-body]
        """
        :returns:
            An {IArray} -- which is an IObjectWithQuantity with the domain for this curve
        """

    @overload
    def __getitem__(self, index: int) -> Tuple[float, float]:
        ...

    @overload
    def __getitem__(self, index: slice) -> Tuple[ValuesType, ValuesType]:
        ...

    def __getitem__(self, index: Union[int, slice]) -> Any:
        """
        [] operator, supporting slices.

        :param index:
            The index of the curve to return

        :returns:
            Returns the pair (domain, image) for the passed slice.
        """



from pytest import approx

from barril import units
from barril.units import ObtainQuantity
from barril.units.posc import CreateAreaQuantityFromLengthQuantity
from barril.units.posc import CreateVolumeQuantityFromLengthQuantity
from barril.units.unit_database import UnitDatabase


def testPoscFrequency(unit_database_posc) -> None:
    u1 = units.Scalar("frequency", 1, "Hz")
    assert u1.GetValue("GHz") == 1.0e-9

    u1 = units.Scalar("frequency", 1, "GHz")
    assert u1.GetValue("Hz") == 1e9


def testPoscTime(unit_database_posc) -> None:
    """
    These are the time-units available on posc.
    """
    units.Scalar("time", 100, "s")
    units.Scalar("time", 100, "min")
    hs = units.Scalar("time", 24, "h")
    d = units.Scalar("time", 1, "d")
    units.Scalar("time", 100, "wk")

    assert hs.GetValue("d") == d.GetValue("d")


def testPoscTemperature(unit_database_posc) -> None:
    u1 = units.Scalar("thermodynamic temperature", 100, "degC")
    u2 = units.Scalar("thermodynamic temperature", 100, "degF")
    u3 = units.Scalar("thermodynamic temperature", 0, "degC")
    u4 = units.Scalar("thermodynamic temperature", 0, "degC")
    u5 = units.Scalar("thermodynamic temperature", 235, "degF")
    u6 = units.Scalar("thermodynamic temperature", 64, "degC")
    assert u1.GetQuantityType() == u2.GetQuantityType()
    assert "temperature" == u2.GetQuantityType()
    assert u1.unit != u2.unit
    assert u3 == u4
    # u1.unit = 'K'  # from C to K
    assert approx(abs(u1.GetValue("K") - 373.15), 7) == 0
    # u2.unit = 'K'  # from F to K
    assert approx(abs(u2.GetValue("K") - 310.927777777), 7) == 0
    # u3.unit = 'degF'  # from C to F
    assert u3.GetValue("degF") == 32.0
    # C to F, F to C
    assert approx(abs(u5.GetValue("degC") - 112.7777777777), 7) == 0
    assert approx(abs(u6.GetValue("degF") - 147.2), 7) == 0
    # now return u3.unit from F to C and compare u3.value with u4.value
    # sanity-check
    assert u3.GetValue("degC") == u4.value


def testPoscRankine(unit_database_posc) -> None:
    rankine = [0.0, 100, 550, 1300]
    kelvin = [0.0, 55.55555, 305.55555, 722.22222]
    unit_database = unit_database_posc

    obtained = unit_database.Convert("thermodynamic temperature", "degR", "K", rankine)
    assert approx(obtained) == kelvin


def testPoscLength(unit_database_posc) -> None:
    # 'length'
    u1 = units.Scalar("length", 10, "ft")
    u2 = units.Scalar("length", 100, "km")
    u3 = units.Scalar("length", 15, "yd")
    u4 = units.Scalar("length", 15, "yd")
    assert u1.GetQuantityType() == u2.GetQuantityType()
    assert u1.unit != u2.unit
    assert u3 == u4
    # u1.unit = 'm'  # from feet to metres
    assert approx(abs(u1.GetValue("m") - 3.048), 7) == 0
    # u2.unit = 'm'  # from kilometres to metres
    assert approx(abs(u2.GetValue("m") - 100000.0), 7) == 0
    # u3.unit = 'ft'  # from yd to ft
    assert approx(abs(u3.GetValue("ft") - 45.0), 7) == 0
    # now return u3.unit from feet to yards and compare u3.value with u4.value
    # sanity-check
    assert u3.GetValue("yd") == u4.value


def testPoscVolume(unit_database_posc) -> None:
    million_cubic_meters = units.Scalar("volume", 1, "M(m3)")
    assert million_cubic_meters.value == 1.0
    assert million_cubic_meters.GetValue("m3") == 1.0e6
    assert million_cubic_meters.GetValue("1000m3") == 1.0e3

    cubic_meters = units.Scalar("volume", 1.0e6, "m3")
    assert cubic_meters.value == 1.0e6
    assert cubic_meters.GetValue("M(m3)") == 1.0
    assert cubic_meters.GetValue("1000m3") == 1.0e3


def testPoscVolumeFlowRate(unit_database_posc) -> None:
    million_cubic_meters = units.Scalar("volume flow rate", 1, "M(m3)/d")
    assert million_cubic_meters.value == 1.0
    assert million_cubic_meters.GetValue("m3/d") == 1.0e6
    assert million_cubic_meters.GetValue("1000m3/d") == 1.0e3

    cubic_meters = units.Scalar("volume flow rate", 1.0e6, "m3/d")
    assert cubic_meters.value == 1.0e6
    assert cubic_meters.GetValue("M(m3)/d") == 1.0
    assert cubic_meters.GetValue("1000m3/d") == 1.0e3


def testPoscPermeabilityLength(unit_database_posc) -> None:
    unit_database = UnitDatabase.GetSingleton()
    assert "volume" == unit_database.GetQuantityType("mD.ft")
    assert "permeability length" == unit_database.GetDefaultCategory("mD.ft")


def testPoscPermeability(unit_database_posc) -> None:
    # 'length'
    u1 = units.Scalar("permeability rock", 1, "D")
    u2 = units.Scalar("permeability rock", 1000, "mD")
    assert u1.GetQuantityType() == u2.GetQuantityType()
    assert u1.unit != u2.unit
    # u1.unit = 'mD'  # from darcy to milidarcy
    # u2.unit = 'D'  # from milidarcy to darcy

    assert approx(abs(u1.GetValue("mD") - 1000.0), 7) == 0
    assert approx(abs(u2.GetValue("D") - 1.0), 7) == 0


def testPoscMolePerMass(unit_database_posc) -> None:
    default = units.Scalar(1, "mol/kg")
    assert default.GetQuantityType() == "mole per mass"

    assert approx(abs(default.GetValue("mol/kg") - 1.0), 7) == 0


def testPoscMolePerTime(unit_database_posc) -> None:
    default = units.Scalar(1, "mol/s")
    assert default.GetQuantityType() == "mole per time"

    assert approx(abs(default.value - 1.0), 7) == 0
    assert approx(abs(default.GetValue("kmol/d") - 1.0 * 86400.0 / 1000.0), 7) == 0
    assert approx(abs(default.GetValue("mol/d") - 1.0 * 86400.0), 7) == 0


def testPoscRotationalFrequency(unit_database_posc) -> None:
    default = units.Scalar(1, "rad/s")
    assert default.GetQuantityType() == "frequency"

    assert approx(abs(default.value - 1.0), 7) == 0
    assert approx(abs(default.GetValue("dega/s") - 57.29578778556937), 7) == 0
    assert approx(abs(default.GetValue("rev/s") - 0.15915494309644432), 7) == 0


def testPoscAngularAcceleration(unit_database_posc) -> None:
    default = units.Scalar(1, "rad/s2")
    assert default.GetQuantityType() == "angular acceleration"

    assert approx(abs(default.value - 1.0), 7) == 0
    assert approx(abs(default.GetValue("dega/s2") - 57.29578778556937), 7) == 0
    assert approx(abs(default.GetValue("dega/min2") - (1.0 * 3600.0) * 57.29578778556937), 7) == 0

    assert approx(abs(default.GetValue("rev/s2") - 0.15915494309644432), 7) == 0
    assert approx(abs(default.GetValue("rev/min2") - 0.15915494309644432 * 3600.0), 7) == 0
    assert approx(abs(default.GetValue("rpm/s") - 9.549296585786658), 7) == 0


def testPoscDensity(unit_database_posc) -> None:
    default = units.Scalar(1, "kg/m3")
    assert default.GetQuantityType() == "density"

    assert approx(abs(default.value - 1.0), 7) == 0
    assert approx(abs(default.GetValue("mg/m3") - 1e6), 7) == 0
    assert approx(abs(default.GetValue("mg/cm3") - 1e6 / 1e6), 7) == 0


def testPoscSpecificEnergy(unit_database_posc) -> None:
    default = units.Scalar(1, "J/kg")
    assert default.GetQuantityType() == "specific energy"

    assert approx(abs(default.value - 1.0), 7) == 0
    assert approx(abs(default.GetValue("J/g") - 1.0 / 1e3), 7) == 0
    assert approx(abs(default.GetValue("kW.h/kg") - 2.7777777777777776e-07), 7) == 0
    assert approx(abs(default.GetValue("kW.h/t") - 2.7777777777777776e-07 * 1e3), 7) == 0

    assert approx(abs(default.GetValue("kW.h/tonUS") - 3.06197599869e-10), 20) == 0
    assert approx(abs(default.GetValue("kW.h/tonUK") - 2.73390677574e-10), 20) == 0


def testPoscMassPerEnergy(unit_database_posc) -> None:
    default = units.Scalar(1, "kg/J")
    assert default.GetQuantityType() == "mass per energy"

    assert approx(abs(default.value - 1.0), 7) == 0
    assert approx(abs(default.GetValue("lbm/Btu") - 1.0 / 0.0004299226139295), 7) == 0


def testKinematicViscosity(unit_database_posc) -> None:
    default = units.Scalar("kinematic viscosity", 1, "m2/s")
    assert default.GetQuantityType() == "volume per time per length"
    assert approx(abs(default.value - 1.0), 7) == 0
    assert approx(abs(default.GetValue("cm2/s") - 10000.0), 7) == 0
    assert approx(abs(default.GetValue("St") - 10000.0), 7) == 0
    assert approx(abs(default.GetValue("cSt") - 1000000.0), 7) == 0
    assert approx(abs(default.GetValue("ft2/h") - 318750.077500155), 7) == 0
    assert approx(abs(default.GetValue("ft2/s") - 10.763910416709722), 7) == 0
    assert approx(abs(default.GetValue("in2/s") - 1550.0031000062002), 7) == 0
    assert approx(abs(default.GetValue("m2/h") - 3600.0), 7) == 0
    assert approx(abs(default.GetValue("mm2/s") - 1000000.0), 7) == 0
    assert approx(abs(default.GetValue("m2/d") - 86400.0), 7) == 0
    assert approx(abs(default.GetValue("ft2/d") - 930001.8600037199), 7) == 0


def testCreateVolumeQuantityFromLengthQuantity(unit_database_posc) -> None:
    unit_database = unit_database_posc
    length_units = unit_database.GetValidUnits("length")
    convertable_length_units = ["m", "cm", "dm", "ft", "in", "km", "mi", "mm", "yd", "um"]

    for length_unit in length_units:
        length_quantity = ObtainQuantity(length_unit, "length")
        volume_quantity = CreateVolumeQuantityFromLengthQuantity(length_quantity)

        if length_unit in convertable_length_units:
            expected_volume_category = "volume"

        else:
            expected_volume_category = "(length) ** 3"

        assert volume_quantity.GetCategory() == expected_volume_category
        assert volume_quantity.GetUnit() == "%s3" % length_unit


def testCreateAreaQuantityFromLengthQuantity(unit_database_posc) -> None:
    unit_database = unit_database_posc
    length_units = unit_database.GetValidUnits("length")
    convertable_length_units = ["m", "cm", "ft", "in", "km", "mi", "miUS", "mm", "um", "yd"]

    for length_unit in length_units:
        length_quantity = ObtainQuantity(length_unit, "length")
        area_quantity = CreateAreaQuantityFromLengthQuantity(length_quantity)

        if length_unit in convertable_length_units:
            expected_volume_category = "area"

        else:
            expected_volume_category = "(length) ** 2"

        assert area_quantity.GetCategory() == expected_volume_category
        assert area_quantity.GetUnit() == "%s2" % length_unit


def testDefaultCategories(unit_database_posc) -> None:
    """
    Check all units with categories defined
    """
    unit_database = unit_database_posc
    categories = set()
    for unit in unit_database.GetUnits():
        category = unit_database.GetDefaultCategory(unit)
        assert category is not None, "All units MUST have a default category."
        quantity_type = unit_database.GetQuantityType(unit)

        if category != quantity_type:
            categories.add(category)

    assert sorted(categories) == [
        "amount of substance",
        "angle per time",
        "area per volume",
        "concentration",
        "delta temperature",
        "dynamic viscosity",
        "energy length per area",
        "energy length per time area temperature",
        "energy per area",
        "energy per length",
        "fluid gas concentration",
        "force length per length",
        "force per force",
        "heat flow rate",
        "index",
        "length per volume",
        "linear thermal expansion",
        "luminous exitance",
        "mass concentration",
        "mass per length",
        "mobility",
        "mole per mole",
        "multiplier",
        "operations per time",
        "permeability length",
        "permeability rock",
        "pressure squared per (dynamic viscosity)",
        "relative power",
        "relative time",
        "self inductance per length",
        "shear rate",
        "status",
        "volume per area",
        "volume per length",
        "volume per time per area",
        "volume per time per volume",
        "volume per volume",
    ]


def testPoscValidUnitsNotRepeated(unit_database_posc) -> None:
    unit_database = unit_database_posc
    for category in unit_database.IterCategories():
        valid_units = unit_database.GetValidUnits(category)
        assert len(valid_units) == len(
            set(valid_units)
        ), f'There is a duplicate unit defined in "{category}": {valid_units}'


def testPoscKVmm(unit_database_posc) -> None:
    """
    V/m to KV/mm (multiply by 1e6)
    """
    q = ObtainQuantity("V/m")
    assert q.ConvertScalarValue(1, "KV/mm") == 1e6


def testPoscBytes(unit_database_posc) -> None:
    """
    Bytes, kBytes, MBytes, ....
    """
    q = ObtainQuantity("Byte")
    assert q.ConvertScalarValue(1024, "kByte") == 1
    assert q.ConvertScalarValue(1024 * 1024, "MByte") == 1
    assert q.ConvertScalarValue(1024 * 1024 * 1024, "GByte") == 1
    assert q.ConvertScalarValue(1024 * 1024 * 1024 * 1024, "TByte") == 1


def testOhmUnits(unit_database_posc) -> None:
    q = ObtainQuantity("ohm/m")
    assert q.ConvertScalarValue(1, "ohm/km") == 1000


def testPower(unit_database_posc) -> None:
    unit_database = unit_database_posc
    assert "volt ampere" == ObtainQuantity("VA").GetQuantityType()
    assert "volt ampere reactive" == ObtainQuantity("VAr").GetQuantityType()

    assert unit_database.Convert("force", "N", "kN", 56.0) == 0.056
    assert (
        approx(
            abs(
                unit_database.Convert("force", [("N", 2)], [("kN", 2)], -(56.0 * 56))
                - -(0.056 * 0.056)
            ),
            7,
        )
        == 0
    )


def testFluidGasConcentration(unit_database_posc) -> None:
    """
    Total Gas Unit requested for PWDA as concentration
    """
    q = ObtainQuantity("tgu")

    assert "dimensionless" == q.GetQuantityType()

    assert approx(abs(q.ConvertScalarValue(1, "ppm") - 333.33), 7) == 0
    assert approx(abs(q.ConvertScalarValue(1, "%") - 0.033333), 7) == 0
    assert approx(abs(q.ConvertScalarValue(1, "Euc") - 0.00033333), 7) == 0


def testSpringDashpotUnits(unit_database_posc) -> None:
    """
    Units used to define Spring-Dashpot movements
    """

    q = ObtainQuantity("Ns/m")
    assert "force per velocity" == q.GetQuantityType()
    assert approx(abs(q.ConvertScalarValue(1, "lbf.s/ft") - 14.5939029372), 7) == 0

    q = ObtainQuantity("Nm/rad")
    assert "moment per angle" == q.GetQuantityType()
    assert approx(abs(q.ConvertScalarValue(1, "lbf.ft/dega") - 836.169044926), 7) == 0

    q = ObtainQuantity("Nms/rad")
    assert "moment per angular velocity" == q.GetQuantityType()
    assert approx(abs(q.ConvertScalarValue(1, "lbf.ft.s/dega") - 0.017453292519943), 7) == 0


def testConcentrationRatio(unit_database_posc) -> None:
    """
    Units used to express a ratio of concentrations
    """
    concentration_ratio_units = unit_database_posc.GetValidUnits("concentration ratio")
    assert concentration_ratio_units == ["mg/l/mg/l", "kg/m3/kg/m3"]


def testJouleThomsonCoefficientUnits() -> None:
    """
    Results gathered from NIST.
    """
    joule_thomson_si_units = units.Scalar("joule-thomson coefficient", 1, "K/Pa")
    assert joule_thomson_si_units.value == 1.0
    assert joule_thomson_si_units.GetValue("degC/Pa") == 1.0
    assert joule_thomson_si_units.GetValue("degC/MPa") == 1e-6
    assert joule_thomson_si_units.GetValue("degC/bar") == 1e-5
    assert joule_thomson_si_units.GetValue("K/MPa") == 1e-6
    assert joule_thomson_si_units.GetValue("K/bar") == 1e-5
    assert joule_thomson_si_units.GetValue("degF/Pa") == approx(9 / 5)
    assert joule_thomson_si_units.GetValue("degF/MPa") == approx(9 / 5 * 1e-6)
    assert joule_thomson_si_units.GetValue("degF/bar") == approx(9 / 5 * 1e-5)
    assert joule_thomson_si_units.GetValue("degR/Pa") == approx(9 / 5)
    assert joule_thomson_si_units.GetValue("degR/MPa") == approx(9 / 5 * 1e-6)
    assert joule_thomson_si_units.GetValue("degR/bar") == approx(9 / 5 * 1e-5)

    joule_thomson_si_units_mega = units.Scalar("joule-thomson coefficient", 1, "K/MPa")
    assert joule_thomson_si_units_mega.GetValue("K/Pa") == 1e6
    assert joule_thomson_si_units_mega.GetValue("K/bar") == 1e1
    assert joule_thomson_si_units_mega.GetValue("degC/Pa") == 1e6
    assert joule_thomson_si_units_mega.GetValue("degC/bar") == 1e1
    assert joule_thomson_si_units_mega.GetValue("degF/Pa") == approx(9 / 5 * 1e6)
    assert joule_thomson_si_units_mega.GetValue("degF/bar") == approx(9 / 5 * 1e1)


def testDensityDerivativePerTemperatureUnitConversion() -> None:
    density_derivative_per_temperature = units.Scalar(
        "density derivative in respect to temperature", 1, "kg/m3.degC"
    )
    assert density_derivative_per_temperature.value == 1.0
    assert density_derivative_per_temperature.GetValue("kg/m3.K") == 1.0

    density_derivative_per_temperature = units.Scalar(
        "density derivative in respect to temperature", 1, "kg/m3.K"
    )
    assert density_derivative_per_temperature.GetValue("kg/m3.degC") == 1.0


def testStandardVolumePerStandardVolumeSmoke() -> None:
    standard_per_standard = units.Scalar(
        "standard volume per standard volume", 1, "scm(15C)/scm(15C)"
    )
    assert standard_per_standard.GetValue("scm(15C)/scm(15C)") == approx(1.0)
    assert standard_per_standard.GetValue("scf(60F)/stb") == approx(5.625408383313)
    assert standard_per_standard.GetValue("scf(60F)/scf") == approx(1.001928010069)
    assert standard_per_standard.GetValue("stb(60F)/stb") == approx(1.0)
    assert standard_per_standard.GetValue("sm3/sm3") == approx(1.0)
    assert standard_per_standard.GetValue("scm3/scm3") == approx(1.0)
    assert standard_per_standard.GetValue("stb/stb") == approx(1.0)
    assert standard_per_standard.GetValue("Mscf/stb") == approx(0.0056254083833)
    assert standard_per_standard.GetValue("MMscf/stb") == approx(0.0000056254083833)
    assert standard_per_standard.GetValue("scf/stb") == approx(5.62540838331)
    assert standard_per_standard.GetValue("stb/scf") == approx(0.17776487178535644)
    assert standard_per_standard.GetValue("stb/MMscf") == approx(177764.87178535643)


def testHeatTransferCoefficient() -> None:
    default = units.Scalar("heat transfer coefficient", 1, "W/m2.K")
    assert default.value == 1.0
    assert default.GetValue("Btu/hr.ft2.degF") == approx(0.17611019426187197)
    assert default.GetValue("Btu/hr.ft2.degR") == approx(0.17611019426187197)
    assert default.GetValue("Btu/hr.m2.degC") == approx(3.412141285851795)
    assert default.GetValue("Btu/s.ft2.degF") == approx(4.891949074810131e-05)
    assert default.GetValue("cal/h.cm2.degC") == approx(0.08604208146120104)
    assert default.GetValue("cal/s.cm2.degC") == approx(2.390057361376673e-05)
    assert default.GetValue("J/s.m2.degC") == approx(1.0)
    assert default.GetValue("kcal/h.m2.degC") == approx(0.8604208146120104)
    assert default.GetValue("cal/h.m2.degC") == approx(860.4208146120104)
    assert default.GetValue("kJ/h.m2.K") == approx(3.6)
    assert default.GetValue("kW/m2.K") == approx(0.001)


def testProductivityIndex() -> None:
    default = units.Scalar("productivity index", 1, "m3/Pa.s")
    assert default.value == 1.0
    assert default.GetValue("Mcf/psi.d") == approx(21037191.806292012)
    assert default.GetValue("ft3/psi.d") == approx(21037191806.292012)
    assert default.GetValue("bbl/d.psi") == approx(3746884212.8623204)
    assert default.GetValue("bbl/kPa.d") == approx(543439633.228566)
    assert default.GetValue("bbl/psi.d") == approx(3746884215.280088)
    assert default.GetValue("L/bar.min") == approx(6000000000.0)
    assert default.GetValue("m3/bar.d") == approx(8640000000.0)
    assert default.GetValue("m3/bar.h") == approx(360000000.0)
    assert default.GetValue("m3/bar.min") == approx(6000000.0)
    assert default.GetValue("m3/d.kPa") == approx(86400000.0)
    assert default.GetValue("m3/kPa.d") == approx(86400000.0)
    assert default.GetValue("m3/kPa.h") == approx(3600000.0)
    assert default.GetValue("m3/psi.d") == approx(595707004.8)
    assert default.GetValue("m3/d/kgf/cm2") == approx(8472945600.38827)


def testThermalConductivity() -> None:
    default = units.Scalar("thermal conductivity", 1, "W/m.K")
    assert default.GetValue("Btu/hr.ft.degF") == approx(0.5777892051642799)
    assert default.GetValue("cal/h.cm.degC") == approx(8.604208146120104)
    assert default.GetValue("cal/s.cm.degC") == approx(0.002390057361376673)
    assert default.GetValue("cal/m.h.degC") == approx(860.4208146120104)
    assert default.GetValue("kcal/h.m.degC") == approx(0.8604208146120104)
    assert default.GetValue("Btu/d.ft.degF") == approx(13.867702121758425)
    assert default.GetValue("kJ/d.m.K") == approx(86.4304235090752)
    assert default.GetValue("W/m.degC") == approx(1.0)


def testStandardVolumePerTime() -> None:
    default = units.Scalar("standard volume per time", 1, "scm(15C)/s")
    assert default.GetValue("ksm3/d") == approx(86.4)
    assert default.GetValue("MMscf(60F)/d") == approx(3.0570698685888087)
    assert default.GetValue("MMscm(15C)/d") == approx(0.0864)
    assert default.GetValue("MMstb(60F)/d") == approx(0.5434396332285661)
    assert default.GetValue("Mscf(60F)/d") == approx(3057.0698685888087)
    assert default.GetValue("Mscm(15C)/d") == approx(86.4)
    assert default.GetValue("Mstb(60F)/d") == approx(543.4396332285661)
    assert default.GetValue("scf(60F)/d") == approx(3057069.868588809)
    assert default.GetValue("scm(15C)/d") == approx(86400.0)
    assert default.GetValue("stb(60F)/d") == approx(543439.6332285661)
    assert default.GetValue("MMscf/d") == approx(3.0570698685888087)
    assert default.GetValue("MMsm3/d") == approx(0.0864)
    assert default.GetValue("MMstb/d") == approx(0.5434396332285661)
    assert default.GetValue("Mscf/d") == approx(3057.069868588809)
    assert default.GetValue("Msm3/d") == approx(0086.4)
    assert default.GetValue("Mstb/d") == approx(543.4396332285661)
    assert default.GetValue("scf/d") == approx(3057069.868588809)
    assert default.GetValue("sm3/d") == approx(86400.0)
    assert default.GetValue("stb/d") == approx(543439.6332285661)
    assert default.GetValue("sm3/s") == approx(1.0)



from collections import OrderedDict
from typing import List

import pytest
from pytest import approx

from barril import units
from barril.units import InvalidUnitError
from barril.units import ObtainQuantity
from barril.units import Quantity


def testFormatting(unit_database_start_units) -> None:
    point = units.FixedArray(2, "length", [(100, 150), (50, 50)], "m")
    assert "(100, 150) (50, 50) [m]" == str(point)


def testValues(unit_database_start_units) -> None:
    with pytest.raises(ValueError):
        units.FixedArray(1, "length", [100], "m")
    point = units.FixedArray(3, "length", [100, 150, 200], "m")
    assert point.unit == "m"
    assert point.values == [100, 150, 200]
    assert point.dimension == 3

    # check dimension read-only
    def SetDimensions():
        point.dimension = 4

    with pytest.raises(AttributeError):
        SetDimensions()

    # check number of values
    with pytest.raises(ValueError):
        point.CreateCopy(values=[0, 0])
    with pytest.raises(ValueError):
        point.CreateCopy(values=[0, 0, 0, 0])
    point = point.CreateCopy(values=[50, 80, 90])  # OK

    # check conversion
    assert point.GetValues("m") == [50, 80, 90]

    assert approx(point.GetValues("km")) == [0.050, 0.080, 0.090]

    assert "50 80 90 [m]" == str(point)

    with pytest.raises(AttributeError):
        setattr(point, "values", [50, 80, 90])


def testInvalidUnits(unit_database_start_units) -> None:
    point = units.FixedArray(2, "length", [(100, 150), (50, 50)], "m")
    try:
        point = point.CreateCopy(values=[(20, 30), (40, 50)], unit="foo")
    except InvalidUnitError:
        pass

    assert point.values == [(100, 150), (50, 50)]
    assert point.unit == "m"


def testCreateWithQuantity(unit_database_start_units) -> None:
    units.FixedArray.CreateWithQuantity(
        Quantity.CreateDerived(OrderedDict()), [100, 150, 200], dimension=3
    )

    quantity = ObtainQuantity("m", "length")
    a1 = units.FixedArray(3, quantity, values=[1, 2, 3])
    a2 = units.FixedArray(3, quantity, values=[1, 2, 3])

    assert a1.GetValues("km") == [0.001, 0.002, 0.003]
    assert a2.GetValues() == [1, 2, 3]


def testCopy(unit_database_start_units) -> None:
    point = units.FixedArray(3, "length", [100, 150, 200], "m")
    assert point.Copy() == point

    #         other = units.FixedArray(3, 'length', [1, 2, 3], 'km')
    other = point.CreateCopy()
    assert other == point

    # test copy with subclasses
    class MyFixedArray(units.FixedArray):
        pass

    point = MyFixedArray(3, "length", [100, 150, 200], "m")
    assert isinstance(point.Copy(), MyFixedArray)

    point = units.FixedArray(3, "length", [100, 150, 200], "m")


def testEmptyArray(unit_database_start_units) -> None:
    arr = units.FixedArray.CreateEmptyArray(3)
    assert not arr.HasCategory()
    assert arr.category == ""
    assert arr.unit == ""

    arr = arr.CreateCopy(category="length", unit="m")
    assert arr.HasCategory()
    assert arr.category == "length"
    assert arr.unit == "m"


def testReadOnlyQuantity(unit_database_start_units) -> None:
    quantity = ObtainQuantity("m", "length")
    array = units.FixedArray(3, quantity, values=[1, 2, 3])
    assert approx(array.GetValues("km")) == [0.001, 0.002, 0.003]


def testDefaultValues(unit_database_start_units) -> None:
    # Not raises exception because by default validation is False on the copy operation
    array: units.FixedArray[List[float]] = units.FixedArray(3, "length")
    assert array.values == [0.0, 0.0, 0.0]

    array = units.FixedArray(3, ObtainQuantity("m"))
    assert array.values == [0.0, 0.0, 0.0]

    with pytest.raises(AssertionError):
        units.FixedArray(3, ObtainQuantity("m"), unit="m")  # type:ignore[call-overload]


def testTupleCreation(unit_database_start_units) -> None:
    """
    Make sure we can use tuples as values.
    """
    array = units.FixedArray(3, (0.0, 200.0, 300.0), "m")
    assert units.FixedArray(3, "length", (0.0, 200.0, 300.0), "m") == array


def testFixedArrayPickle(unit_database_start_units) -> None:
    import pickle

    fixed_array_1 = units.FixedArray(3, "length", values=[1, 2, 3], unit="m")
    fixed_array_2 = pickle.loads(pickle.dumps(fixed_array_1))

    assert fixed_array_1 == fixed_array_2


def testNumberInteractions() -> None:
    # Operations on FixedArrays used to return Array instances.
    a = units.FixedArray(3, [1, 2, 3], "m")
    b = units.FixedArray(3, [0.5] * 3, "m")

    s = a + b
    assert isinstance(s, units.FixedArray)
    assert s.GetDimension() == 3

    m = a + b
    assert isinstance(m, units.FixedArray)
    assert m.GetDimension() == 3


def testFixedArrayChangingIndex() -> None:
    fixed_array = units.FixedArray(3, [1, 2, 3], "m")
    assert fixed_array.ChangingIndex(0, 5) == units.FixedArray(3, [5, 2, 3], "m")
    assert fixed_array.ChangingIndex(0, units.Scalar(5, "m")) == units.FixedArray(3, [5, 2, 3], "m")
    assert fixed_array.ChangingIndex(0, (5,)) == units.FixedArray(3, [5, 2, 3], "m")

    # Different unit
    assert fixed_array.ChangingIndex(0, (5, "cm")) == units.FixedArray(3, [5, 200, 300], "cm")
    assert fixed_array.ChangingIndex(0, units.Scalar(5, "cm")) == units.FixedArray(
        3, [5, 200, 300], "cm"
    )

    # Keeping fixed array unit
    assert fixed_array.ChangingIndex(
        0, units.Scalar(5, "cm"), use_value_unit=False
    ) == units.FixedArray(3, [0.05, 2, 3], "m")

    # Other indexes
    assert fixed_array.ChangingIndex(1, units.Scalar(5, "cm")) == units.FixedArray(
        3, [100, 5, 300], "cm"
    )
    assert fixed_array.ChangingIndex(2, units.Scalar(5, "cm")) == units.FixedArray(
        3, [100, 200, 5], "cm"
    )


def testFixedArrayIndexAsScalar() -> None:
    from barril.units import Scalar

    fixed_array = units.FixedArray(3, "length of path", [1, 2, 3], "m")
    assert fixed_array.IndexAsScalar(0).GetCategory() == "length of path"
    assert fixed_array.IndexAsScalar(0) == Scalar("length of path", 1, "m")
    assert fixed_array.IndexAsScalar(1) == Scalar("length of path", 2, "m")



import pytest
from pytest import approx
from pytest import raises

from barril.units import Scalar
from barril.units.unit_database import _LEGACY_TO_CURRENT


def testCubicFeetPerDayLegacyUnits() -> None:
    # test creating scalar using legacy representation and default category
    q = Scalar(1.0, "1000ft3/d")
    assert q.GetUnit() == "Mcf/d"

    # test creating scalar using legacy representation
    q = Scalar(category="volume flow rate", value=1.0, unit="1000ft3/d")
    assert q.GetUnit() == "Mcf/d"

    # test getting value using legacy representation
    assert approx(q.GetValue("M(ft3)/d")) == q.GetValue("MMcf/d")

    q = Scalar(category="volume flow rate", value=1.0, unit="M(ft3)/d")
    assert q.GetUnit() == "MMcf/d"
    assert approx(q.GetValue("M(ft3)/d")) == q.GetValue("MMcf/d")
    assert q.GetUnitName() == "million cubic feet per day"


@pytest.mark.parametrize("legacy, current", _LEGACY_TO_CURRENT)
@pytest.mark.parametrize("value", [1.0, 3.1415, 123.567])
def testAllLegacyUnits(legacy: str, current: str, value: float) -> None:
    test_scalar = Scalar(value, legacy)
    assert test_scalar.GetUnit() == current
    assert approx(test_scalar.GetValue(legacy)) == value
    assert approx(test_scalar.GetValue(current)) == value


def testCreateScalarUnitsError() -> None:
    from barril.units.unit_database import UnitsError

    with raises(UnitsError, match="Unable to get default category for: foo/d"):
        _ = Scalar(1.0, "foo/d")


def testGetValueInvalidUnitError() -> None:
    from barril.units.unit_database import InvalidUnitError

    with raises(
        InvalidUnitError, match="Invalid unit for quantity_type volume flow rate: 1000ft3/foo"
    ):
        q = Scalar(1.0, "1000ft3/d")
        q.GetValue("1000ft3/foo")


def testFixUnitIfIsLegacyExcept() -> None:
    from barril.units.unit_database import FixUnitIfIsLegacy

    class SomeNonExpectedObject:
        pass

    unknown = SomeNonExpectedObject()
    is_legacy, unit = FixUnitIfIsLegacy(unknown)  # type:ignore[arg-type]
    assert not is_legacy
    assert unit is unknown  # type:ignore[comparison-overlap]


def testUnitDatabaseConvert(unit_database_posc) -> None:
    import numpy as np

    # Test against float
    converted = unit_database_posc.Convert("volume flow rate", "1000ft3/d", "m3/d", 1.0)
    assert approx(converted) == 28.31685

    converted = unit_database_posc.Convert("volume flow rate", "M(ft3)/d", "m3/d", 1.0)
    assert approx(converted) == 28316.85

    # Test against numpy arrays
    converted = unit_database_posc.Convert("volume flow rate", "1000ft3/d", "m3/d", np.ones(2))
    assert approx(converted) == 28.31685
    converted = unit_database_posc.Convert("volume flow rate", "M(ft3)/d", "m3/d", np.ones(2))
    assert approx(converted) == 28316.85


def testUnitDatabaseGetUnitNameLegacy(unit_database_posc) -> None:
    assert (
        unit_database_posc.GetUnitName(quantity_type="volume flow rate", unit="M(ft3)/d")
        == "million cubic feet per day"
    )


def testUnitDatabaseCheckQuantityTypeUnitLegacy(unit_database_posc) -> None:
    from barril.units.unit_database import InvalidUnitError

    with raises(InvalidUnitError, match="Invalid unit for quantity_type volume flow rate"):
        unit_database_posc.CheckQuantityTypeUnit(quantity_type="volume flow rate", unit="M(ft3)/d")


def testUnitDatabaseGetDefaultCategory(unit_database_posc) -> None:
    category = unit_database_posc.GetDefaultCategory("1000ft3/d")
    assert category == "volume flow rate"

    class SomeNonExpectedObject:
        pass

    category = unit_database_posc.GetDefaultCategory(SomeNonExpectedObject())
    assert category is None



import weakref
from typing import Dict

import pytest

from barril.units import Array
from barril.units import FixedArray
from barril.units import FractionScalar
from barril.units import ObtainQuantity
from barril.units import Scalar
from barril.units.unit_system import UnitSystem
from barril.units.unit_system_manager import InvalidTemplateError
from barril.units.unit_system_manager import UnitSystemCategoriesError
from barril.units.unit_system_manager import UnitSystemIDError
from barril.units.unit_system_manager import UnitSystemManager


@pytest.fixture
def unit_manager():
    manager = UnitSystemManager()
    UnitSystemManager.PushSingleton(manager)

    yield manager

    UnitSystemManager.PopSingleton()


@pytest.fixture
def units_mapping_1():
    units_mapping = {"length": "m"}
    return units_mapping


@pytest.fixture
def units_mapping_2():
    units_mapping = {"length": "km"}
    return units_mapping


@pytest.fixture
def units_mapping_3():
    units_mapping: Dict[str, str] = {}
    return units_mapping


def CreateUnitSystemTemplate(manager):
    units_mapping_template = {"length": "m"}
    manager.SetTemplateUnitSystemByUnitsMapping(units_mapping_template)


def testUnitSystemManager(unit_manager, units_mapping_1, units_mapping_2, units_mapping_3) -> None:
    CreateUnitSystemTemplate(unit_manager)
    unit_manager.AddUnitSystem("system1", "system1", units_mapping_1, False)
    unit_manager.AddUnitSystem("system2", "system2", units_mapping_2, False)

    with pytest.raises(UnitSystemIDError):
        unit_manager.AddUnitSystem("system1", "system3", units_mapping_3, False)

    with pytest.raises(UnitSystemCategoriesError):
        unit_manager.AddUnitSystem("system3", "system3", units_mapping_3, False)

    scalar = Scalar("length", unit="cm")
    array = Array("length", unit="cm")
    fixed_array = FixedArray(3, "length", unit="cm")
    fraction_scalar = FractionScalar("length")

    unit_systems = unit_manager.GetUnitSystems()
    system1 = unit_systems["system1"]
    system2 = unit_systems["system2"]

    unit_manager.current = system1
    assert unit_manager.GetUnitSystemById("system1") is system1
    assert unit_manager.GetUnitSystemById("system2") is system2

    # make sure the unit system manager is not holding a strong ref to its objects
    scalar_ref = weakref.ref(scalar)
    array_ref = weakref.ref(array)
    fixed_array_ref = weakref.ref(fixed_array)
    fraction_scalar_ref = weakref.ref(fraction_scalar)

    del scalar
    del array
    del fixed_array
    del fraction_scalar

    assert scalar_ref() is None
    assert array_ref() is None
    assert fixed_array_ref() is None
    assert fraction_scalar_ref() is None


def testGetNewId(unit_manager, units_mapping_1) -> None:
    CreateUnitSystemTemplate(unit_manager)

    new_id = unit_manager.GetNewId()
    assert new_id == "system 1"
    unit_manager.AddUnitSystem(new_id, "system 1", units_mapping_1, False)

    assert unit_manager.GetNewId() == "system 2"
    unit_manager.RemoveUnitSystem("system 1")
    assert unit_manager.GetNewId() == "system 1"


def testCurrentUnitSystemUpdate(unit_manager, units_mapping_1) -> None:
    """
    If the current unit system is removed, another should be set
    """
    CreateUnitSystemTemplate(unit_manager)

    current = unit_manager.GetCurrent()
    assert current.GetId() is None

    # Adding
    unit_manager.AddUnitSystem("system 1", "system 1", units_mapping_1, False)

    # Since this is the first valid unit system we expect it to be set as current
    current_id = unit_manager.current.GetId()
    assert current_id == "system 1"

    # Adding another system, the current is not expected to change
    unit_manager.AddUnitSystem("system 2", "system 2", units_mapping_1, False)
    current_id = unit_manager.current.GetId()
    assert current_id == "system 1"

    # Removing the unit system set as current
    unit_manager.RemoveUnitSystem("system 1")

    # The second system should be set as default
    current_id = unit_manager.current.GetId()
    assert current_id == "system 2"

    # Also removing the second unit system
    unit_manager.RemoveUnitSystem("system 2")
    current_id = unit_manager.current.GetId()
    assert current_id is None


def testConvertToCurrent(unit_manager, units_mapping_1, units_mapping_2) -> None:
    CreateUnitSystemTemplate(unit_manager)

    units_mapping_3 = {"length": "cm"}

    system1 = unit_manager.AddUnitSystem("system1", "", units_mapping_1)
    system2 = unit_manager.AddUnitSystem("system2", "", units_mapping_2)
    system3 = unit_manager.AddUnitSystem("system3", "", units_mapping_3)

    def Convert(value, unit):
        return unit_manager.ConvertToCurrent("length", unit, value)

    def assertConversion(v1, v2):
        assert Convert(*v1) == v2
        prev_scalar = Scalar(v1)
        expected_scalar = Scalar(v2)
        obtained_scalar = unit_manager.ConvertScalarToCurrent(prev_scalar)
        assert obtained_scalar == expected_scalar

    # no current unit system: no conversion performed
    unit_manager.current = None
    assertConversion((100.0, "m"), (100.0, "m"))

    # unit system available: convert to current unit
    unit_manager.current = system1  # m
    assertConversion((100.0, "m"), (100.0, "m"))

    unit_manager.current = system2  # km
    assertConversion((100.0, "m"), (0.1, "km"))

    unit_manager.current = system3  # cm
    assertConversion((100.0, "m"), (10000.0, "cm"))


def testCreateUnitSystemWithoutTemplate(unit_manager, units_mapping_1) -> None:
    """
    It should be possible to create unit systems without a units template.
    """
    system1 = unit_manager.AddUnitSystem("system1", "", units_mapping_1)
    assert unit_manager.GetUnitSystemTemplate() is None

    unit_manager.GetCurrent()
    assert system1.GetId() == "system1"

    # However if a template is set then existing unit systems should respect it, and if any category
    # specified in the template is missing, then an error should be raised.
    units_mapping_template = {"pressure": "psi"}
    with pytest.raises(InvalidTemplateError):
        unit_manager.SetTemplateUnitSystemByUnitsMapping(units_mapping_template)


def testGetQuantityDefaultUnit(unit_manager, units_mapping_1) -> None:
    system1 = unit_manager.AddUnitSystem("system1", "", units_mapping_1)

    system1.SetDefaultUnit("length", "cm")
    quantity = ObtainQuantity(category="length", unit="m")

    default_unit = unit_manager.GetQuantityDefaultUnit(quantity)
    assert default_unit == "cm"


def testEmptyUnitSystemManager() -> None:
    """
    When no current unit system is selected the method GetCurrent should return an empty unit
    system.
    """
    unit_manager = UnitSystemManager()
    current = unit_manager.current

    assert current is not None
    assert isinstance(current, UnitSystem)

    assert current.GetId() is None
    assert current.GetCaption() == "Null"
    assert current.IsReadOnly()

    current_mapping = current.GetUnitsMapping()
    assert len(current_mapping) == 0



from typing import Any
from typing import Dict
from typing import Optional

from oop_ext.foundation import callback
from oop_ext.foundation.decorators import Implements
from oop_ext.interface import ImplementsInterface
from oop_ext.interface import IsImplementation

from .unit_system_interface import IUnitSystem


@ImplementsInterface(IUnitSystem)
class UnitSystem:
    """
    Default implementation for a Unit System.
    .. see:: IUnitSystem
    """

    @Implements(IUnitSystem.__init__)
    def __init__(
        self,
        id: Optional[str],
        caption: str,
        units_mapping: Dict[str, str],
        read_only: bool = False,
    ) -> None:
        self._id = id
        self._caption = caption
        self._units_mapping = units_mapping
        self._read_only = read_only
        self.on_default_unit: callback.Callback2[str, Optional[str]] = callback.Callback2()

    @Implements(IUnitSystem.GetId)
    def GetId(self) -> Optional[str]:
        return self._id

    @Implements(IUnitSystem.GetCaption)
    def GetCaption(self) -> str:
        return self._caption

    @Implements(IUnitSystem.SetCaption)
    def SetCaption(self, caption: str) -> None:
        self._caption = caption

    @Implements(IUnitSystem.GetUnitsMapping)
    def GetUnitsMapping(self) -> Dict[str, str]:
        return self._units_mapping

    @Implements(IUnitSystem.IsReadOnly)
    def IsReadOnly(self) -> bool:
        return self._read_only

    @Implements(IUnitSystem.SetReadOnly)
    def SetReadOnly(self, read_only: bool) -> None:
        self._read_only = read_only

    # : duplicated signature to comply with Get/Set standard
    GetReadOnly = IsReadOnly
    read_only = property(GetReadOnly, SetReadOnly)

    @Implements(IUnitSystem.GetDefaultUnit)
    def GetDefaultUnit(self, category: str) -> Optional[str]:
        if not category:
            return None

        return self._units_mapping.get(category, None)

    @Implements(IUnitSystem.RemoveCategory)
    def RemoveCategory(self, category: str) -> None:
        try:
            del self._units_mapping[category]
            self.on_default_unit(category, None)
        except KeyError:
            # The category is not in the unit system, so there is nothing to do
            pass

    @Implements(IUnitSystem.SetDefaultUnit)
    def SetDefaultUnit(self, category: str, unit: str) -> None:
        self._units_mapping[category] = unit
        self.on_default_unit(category, unit)

    @Implements(IUnitSystem.__eq__)
    def __eq__(self, other: Any) -> bool:
        if not IsImplementation(other, IUnitSystem):
            return False

        return (
            self.GetId() == other.GetId()
            and self.GetCaption() == other.GetCaption()
            and self.GetUnitsMapping() == other.GetUnitsMapping()
            and self.IsReadOnly() == other.IsReadOnly()
        )



from typing import Any
from typing import overload
from typing import Tuple
from typing import Union

from oop_ext.foundation.decorators import Deprecated
from oop_ext.interface import AssertImplements
from oop_ext.interface import ImplementsInterface

from barril.curve.curve_interface import ICurve
from barril.units.interfaces import IArray
from barril.units.interfaces import ValuesType

__all__ = ["Curve"]


@ImplementsInterface(ICurve)
class Curve:
    """
    The curve is an element that has a domain "x" and an image "f(x)".

    Domain: The domain of a given function is the set of 'input' values for which the function is
    defined -- from Wikipedia

    Image: it is defined as 'image set' the set of values which effectively f(x) takes -- from
    Wikipedia (pt)

    :ivar _image:
        The array with the image for this curve

    :ivar _domain:
        The array with the domain for this curve
    """

    def __init__(self, image: IArray, domain: IArray) -> None:
        """
        :param image:
            This is the image of this curve

        :param domain:
            This is the domain of this curve
        """
        AssertImplements(image, IArray)
        AssertImplements(domain, IArray)

        self._CheckImageAndDomainLength(image, domain)
        self._image = image
        self._domain = domain

    def _CheckImageAndDomainLength(self, image: IArray, domain: IArray) -> None:
        """
        Check if image and domain have different lengths, if is True raises ValueError.

        :param image:
            This is the image of this curve

        :param domain:
            This is the domain of this curve

        :raises ValueError:
            if image and domain have different lengths
        """
        image_length = len(image.GetValues())
        domain_length = len(domain.GetValues())
        if image_length != domain_length:
            msg = (
                "The length of the image ({}) is different from the size of the domain ({})".format(
                    image_length, domain_length
                )
            )
            raise ValueError(msg)

    def GetDomain(self) -> IArray:
        return self._domain

    def SetDomain(self, domain: IArray) -> None:
        """
        Define the curve domain.

        :param domain:
            the domain of this curve
        """
        self._CheckImageAndDomainLength(self._image, domain)
        self._domain = domain

    domain = property(GetDomain, SetDomain)

    def GetLength(self) -> int:
        """
        :returns:
            The length of the curve
        """
        return len(self._image.GetValues())

    def GetImage(self) -> IArray:
        return self._image

    def SetImage(self, image: IArray) -> None:
        """
        Define the curve image.

        :param IArray image:
            The image of this curve
        """
        self._CheckImageAndDomainLength(image, self._domain)
        self._image = image

    image = property(GetImage, SetImage)

    @Deprecated("SetImage")
    def SetValues(self, values: IArray) -> None:
        return self.SetImage(values)

    @Deprecated("GetImage")
    def GetValues(self) -> IArray:
        """
        @see ICurve.GetImage.
        """
        return self.GetImage()

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Curve):
            return False
        return self.GetImage() == other.GetImage() and self.GetDomain() == other.GetDomain()

    @overload
    def __getitem__(self, index: int) -> Tuple[float, float]:
        ...

    @overload
    def __getitem__(self, index: slice) -> Tuple[ValuesType, ValuesType]:
        ...

    def __getitem__(self, index: Union[int, slice]) -> Tuple[Any, Any]:
        d = self.GetDomain().GetValues()[index]  # type:ignore[index]
        i = self.GetImage().GetValues()[index]  # type:ignore[index]
        return d, i

    def __repr__(self) -> str:
        image = self.GetImage()
        domain = self.GetDomain()

        xy = []
        for i, (x, y) in enumerate(zip(image, domain)):  # type:ignore[call-overload]
            if i > 20:
                xy.append(" ... ")
                break
            xy.append(f"({x}, {y})")

        return "Curve({}, {})[{}]".format(
            image.unit, domain.unit, " ".join(xy)  # type:ignore[attr-defined]
        )



import pytest

from barril.curve.curve import Curve
from barril.units import Array
from barril.units import ObtainQuantity
from barril.units import UnitDatabase


@pytest.fixture
def unit_database():
    unit_database = UnitDatabase()
    unit_database.AddUnit("length", "milimeters", "mm", "%f * 1000.0", "%f / 1000.0")
    unit_database.AddUnitBase("length", "meters", "m")
    unit_database.AddUnit("length", "centimeters", "cm", "%f * 100.0", "%f / 100.0")
    unit_database.AddUnit("length", "kilometers", "km", "%f / 1000.0", "%f * 1000.0")

    unit_database.AddUnitBase("time", "seconds", "s")
    unit_database.AddUnit("time", "minutes", "min", "%f * 60.0", " %f * 60.0")
    unit_database.AddUnit("time", "hours", "h", "%f * 3600.0", " %f * 3600.0")
    unit_database.AddUnit("time", "days", "d", "%f * 86400.0", " %f * 86400.0")

    unit_database.AddCategory(
        category="length", quantity_type="length", valid_units=["cm", "m", "km"]
    )
    unit_database.AddCategory(
        category="time", quantity_type="time", valid_units=["s", "min", "h", "d"]
    )

    UnitDatabase.PushSingleton(unit_database)
    yield unit_database

    UnitDatabase.PopSingleton()


def testCurves(unit_database) -> None:
    import numpy

    r = ObtainQuantity("m", "length")

    values10 = Array(r, values=numpy.array(list(range(10)), dtype=numpy.int32))
    values9 = Array(r, values=numpy.array(list(range(9)), dtype=numpy.int32))

    domain9 = Array("time", values=numpy.array(list(range(9)), dtype=numpy.int32), unit="s")
    domain10 = Array("time", values=numpy.array(list(range(10)), dtype=numpy.int32), unit="s")

    with pytest.raises(ValueError):
        Curve(values10, domain9)  # type:ignore[arg-type]

    c = Curve(values10, domain10)  # type:ignore[arg-type]

    with pytest.raises(ValueError):
        c.SetDomain(domain9)  # type:ignore[arg-type]
    with pytest.raises(ValueError):
        c.SetImage(values9)  # type:ignore[arg-type]


def testSlice(unit_database) -> None:
    quantity = ObtainQuantity("m", "length")

    def MakeArray(values):
        return Array(quantity, values=values)

    domain = MakeArray([5, 10, 20, 30, 40, 50])
    image = MakeArray([0, 1, 2, 3, 4, 5])

    curve = Curve(image, domain)

    assert curve[0] == (5, 0)
    assert curve[:3] == ([5, 10, 20], [0, 1, 2])
    assert curve[:] == ([5, 10, 20, 30, 40, 50], [0, 1, 2, 3, 4, 5])


def testCurveRepr(unit_database) -> None:
    q1 = ObtainQuantity("m", "length")
    q2 = ObtainQuantity("d", "time")
    curve = Curve(Array(q1, []), Array(q2, []))  # type:ignore[arg-type]
    assert "Curve(m, d)[]" == repr(curve)

    curve = Curve(Array(q1, list(range(3))), Array(q2, list(range(3))))  # type:ignore[arg-type]
    assert "Curve(m, d)[(0, 0) (1, 1) (2, 2)]" == repr(curve)

    curve = Curve(Array(q1, list(range(100))), Array(q2, list(range(100))))  # type:ignore[arg-type]
    expected = (
        "Curve(m, d)[(0, 0) (1, 1) (2, 2) (3, 3) (4, 4) (5, 5) "
        "(6, 6) (7, 7) (8, 8) (9, 9) (10, 10) (11, 11) "
        "(12, 12) (13, 13) (14, 14) (15, 15) (16, 16) "
        "(17, 17) (18, 18) (19, 19) (20, 20)  ... ]"
    )
    assert repr(curve) == expected



import pytest

from barril._util.types_ import _GetKnownNumberTypes
from barril._util.types_ import CheckFormatString
from barril._util.types_ import CheckType
from barril._util.types_ import IsNumber


def testPassing() -> None:
    class Foo:
        pass

    CheckType(Foo(), Foo)
    CheckType(Foo(), (int, Foo))
    CheckType(99, (int, Foo))


def testRaising() -> None:
    with pytest.raises(TypeError):
        CheckType("hellou", int)

    with pytest.raises(TypeError):
        CheckType("hellou", (int, float))

    with pytest.raises(TypeError):
        CheckType(99, (str, float))


def testCheckFormatString() -> None:
    CheckFormatString("%s", 1)
    CheckFormatString("%s m", 1)

    with pytest.raises(ValueError):
        CheckFormatString("%s m %s", 1)

    with pytest.raises(ValueError):
        CheckFormatString("%s m %s", 1, 3, 3)


def testIfCustomMessageIsAppendedToDefaultMessage() -> None:
    message = "Zero is not unicode!"

    with pytest.raises(TypeError) as exception:
        CheckType(0, str, message)

    assert message in str(exception.value)


def testCheckNumber() -> None:
    numpy = pytest.importorskip("numpy")

    for number_class in (float, int):
        converted = number_class(1)
        assert IsNumber(converted)

    # Checking numpy numbers
    collection = numpy.zeros(1, numpy.float32)
    assert IsNumber(collection[0])


def testGetKnownNumberTypes(monkeypatch) -> None:
    import sys

    numpy = pytest.importorskip("numpy")

    expected = {float, complex, numpy.number}
    expected.update({int})
    assert set(_GetKnownNumberTypes()) == expected

    monkeypatch.setitem(sys.modules, "numpy", None)
    expected.remove(numpy.number)
    assert set(_GetKnownNumberTypes()) == expected



import weakref
from collections import OrderedDict
from copy import deepcopy
from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import Type
from typing import TYPE_CHECKING

from oop_ext.foundation import callback
from oop_ext.foundation.decorators import Override
from oop_ext.foundation.singleton import Singleton
from oop_ext.interface import AssertImplements

from .unit_database import UnitDatabase
from .unit_system import UnitSystem
from .unit_system_interface import IUnitSystem
from barril.units import AbstractValueWithQuantityObject
from barril.units import IQuantity


if TYPE_CHECKING:
    from barril.units import Scalar


class NoTemplateError(RuntimeError):
    """
    Error raised when trying to create a new unit system and no template was defined.
    """

    def __init__(self) -> None:
        msg = "Error creating unit system. There is no template defined."
        RuntimeError.__init__(self, msg)


class InvalidTemplateError(RuntimeError):
    """
    Error raised when a template is set in a manager that has at least one unit system that does not
    match the given template
    """

    def __init__(self, invalid_unit_system_ids: Optional[List[Optional[str]]] = None):
        """
        :param invalid_unit_system_ids:
            The ids of the unit system that do not match a given template
        """
        if invalid_unit_system_ids:
            msg = "The unit systems %s do not match the given template" % str(
                invalid_unit_system_ids
            )
        else:
            msg = "At least one of the current unit systems does not match the given template"
        RuntimeError.__init__(self, msg)


class UnitSystemCategoriesError(KeyError):
    """
    Error raised when the categories defined in a new unit system differ from those define in the
    template.
    """

    def __init__(self, default_categories: List[str], current_categories: List[str]):
        msg = "Error creating new unit system. The default categories are %s but %s were defined."
        KeyError.__init__(self, msg % (default_categories, current_categories))


class UnitSystemIDError(KeyError):
    """
    Error raised when trying to create a new unit system given an already used ID.
    """

    def __init__(self, id: str):
        msg = f"Error creating new unit system. The ID {id} is already in use."
        KeyError.__init__(self, msg)


class UnitSystemManager(Singleton):
    """
    A service that manages the unit systems. It handles the unit system objects creation and
    selection.

    :ivar _object_refs:
        Set of weakref for all objects which represents a value+unit.

    :ivar current:
        The current unit system.

    :ivar on_current:
        Callback called when the current unit system changes (the unit system is given as parameter).

    :ivar on_unit_changed:
        Callback called when the the unit of a category changes (the category and unit is given as
        parameter).

    :ivar _unit_systems:
        A dict that holds the available unit systems.
    """

    def __init__(self) -> None:
        # list with objects with a unit associated with it.
        self._object_refs: Set[_IdentityWrap] = set()

        # someone would want to listen to changes in the current unit system
        self.on_current = callback.Callback1[IUnitSystem]()

        # someone would want to listen to changes in unit system categories unit.
        self.on_unit_changed = callback.Callback2[str, Optional[str]]()

        # the current unit system which is being used by the application
        self._current: Optional[IUnitSystem] = None

        # container for registered unit systems (strong references)
        self._unit_systems: OrderedDict[str, IUnitSystem] = OrderedDict()

        # unit system template.
        self._unit_system_template: Optional[IUnitSystem] = None

        # default unit system class.
        self._default_unit_system_class: Type[IUnitSystem] = UnitSystem

        # This unit system is returned when no current unit system is selected
        self.__null_unit_system = UnitSystem(
            id=None, caption="Null", units_mapping={}, read_only=True
        )

    @Override(Singleton.ResetInstance)
    def ResetInstance(self) -> None:
        self.on_current.UnregisterAll()
        self.on_unit_changed.UnregisterAll()

    def SetDefaultUnitSystemClass(self, class_: Type[IUnitSystem]) -> None:
        """
        Set the default unit system class.

        :param class_:
            The unit system class that will be used when creating new unit systems.
        """
        AssertImplements(class_, IUnitSystem)
        self._default_unit_system_class = class_

    def SetTemplateUnitSystemByUnitsMapping(self, units_mapping: Dict[str, str]) -> None:
        """
        Set the unit system template based on the given units mapping.

        This template will be used to validate new unit systems.

        :param units_mapping:
            A dict that maps each category to a related unit (which will be set as default).
            The valid categories are defined by the coilib50.units.UnitDatabase.

        :raise: TemplateDefinedAfterUnitSystemError
            See TemplateDefinedAfterUnitSystemError documentation.
        """
        template_categories = list(units_mapping.keys())
        invalid_unit_systems = []

        # Check if existing unit systems match the given template
        for unit_system in list(self._unit_systems.values()):
            current_units_mapping = unit_system.GetUnitsMapping()
            match = self._CheckUnitSystemMapping(current_units_mapping, template_categories)

            if not match:
                invalid_unit_systems.append(unit_system.GetId())

        if invalid_unit_systems:
            # At least one of the existing unit system is not valid for this template. Notify that
            # the template is invalid
            raise InvalidTemplateError(invalid_unit_systems)

        # NOTE: 'tr' for the caption (Unit system template) was removed.
        self._unit_system_template = self._default_unit_system_class(
            "template", "Unit system template", units_mapping, True
        )

    def GetUnitSystemTemplate(self) -> Optional[IUnitSystem]:
        """
        :returns:
            Return the template unit system.
        """
        return self._unit_system_template

    def _CheckUnitSystemMapping(
        self, units_mapping: Dict[str, str], required_categories: List[str]
    ) -> bool:
        """
        Checks if the given units mapping have a default unit for all required categories

        :param units_mapping units_mapping:
            A dict that maps each category to a related unit. the dict should contain mapping for
            each category defined in the template.

        :param required_categories:
            The categories required in the given unit system.
        """
        required_categories_set = set(required_categories)

        current_categories = list(units_mapping.keys())

        return set(current_categories).issuperset(required_categories_set)

    def AddUnitSystem(
        self,
        id: str,
        caption: str,
        units_mapping: Optional[Dict[str, str]] = None,
        read_only: bool = False,
    ) -> IUnitSystem:
        """
        Create a new unit system based on the given parameters.

        When creating a unit system it will be kept internally at the manager in a unit systems
        list and set as current right after the creation.

        It is expected that a unit system template is already set. The template will be used to
        validate the categories from the given units mapping.

        :param id:
            The unique id of the unit system.

        :param caption:
            The user-friendly name of the unit system.

        :param units_mapping:
            A dict that maps each category to a related unit (which will be set as default).
            The valid categories are defined by the coilib50.units.UnitDatabase.
            If None is given the units mapping from the template will be taken.

        :param read_only:
            Flag that sets the read-only state of the new unit system.

        :raises NoTemplateError:
            See NoTemplateError documentation.

        :raises UnitSystemIDError:
            See UnitSystemIDError documentation.

        :raises UnitSystemCategoriesError:
            See UnitSystemCategoriesError documentation.

        :returns:
            The created unit system.
        """
        if id in self._unit_systems:
            raise UnitSystemIDError(id)

        if self._unit_system_template is not None:
            template_units_mapping = self._unit_system_template.GetUnitsMapping()

            # If a unit map was passed we need to check it, otherwise just copy the template into
            # the new system
            if units_mapping is None:
                units_mapping = deepcopy(template_units_mapping)

            else:
                template_categories = list(template_units_mapping.keys())
                match = self._CheckUnitSystemMapping(units_mapping, template_categories)

                if not match:
                    raise UnitSystemCategoriesError(template_categories, list(units_mapping.keys()))

        else:
            # No template to check
            if units_mapping is None:
                # If not template was given, and not unit mapping was provided, let us create
                # an empty unit system.
                units_mapping = {}

        unit_system = self._default_unit_system_class(id, caption, units_mapping, read_only)
        self._unit_systems[id] = unit_system

        if self._current is None:
            self.SetCurrent(unit_system)

        return unit_system

    def _CategoryUnitChange(self, category: str, unit: Optional[str]) -> None:
        """
        Triggers on_unit_changed callback when default unit change on a category in a unit system.

        :param category:
            The changed category.

        :param unit:
            The new unit.
        """
        self.on_unit_changed(category, unit)

    def RemoveUnitSystem(self, unit_system_id: str) -> None:
        """
        Remove the unit system with the given id.

        :param unicode unit_system_id:
            The id of the unit system to be removed.
        """
        del self._unit_systems[unit_system_id]

        # If the current unit system was removed, set another system as current
        assert self._current is not None
        if self._current.GetId() == unit_system_id:
            available = list(self.GetUnitSystems().values())

            if available:
                self.SetCurrent(available[0])
            else:
                self.SetCurrent(None)

    def GetUnitSystems(self) -> Dict[str, IUnitSystem]:
        """
        :returns:
            The dict containing all the registered unit systems indexed by their ids.
        """
        return self._unit_systems

    # Current --------------------------------------------------------------------------------------

    def SetCurrent(self, unit_system: Optional[IUnitSystem]) -> None:
        """
        Sets a new current unit system, updating all registered objects. If None, no object is changed.

        :param unit_system:
            The new current unit system.
        """
        if self._current is not None:
            self._current.on_default_unit.Unregister(self._CategoryUnitChange)

        self._current = unit_system

        if self._current is not None:
            self._current.on_default_unit.Register(self._CategoryUnitChange)
            self.on_current(self._current)
        else:
            # Notifying that an empty unit system was set for the clients
            self.on_current(self.__null_unit_system)

        self.UpdateObjects()

    def GetCurrent(self) -> IUnitSystem:
        """
        Retrieves the unit system currently in use.

        :rtype: IUnitSystem instance
        :returns:
            The currently used unit system.
        """
        result = self._current

        if result is None:
            # If no system is set as current, we will return a null unit system just to keep the
            # interface (so that clients do not need to keep checking if the received system is null)
            result = self.__null_unit_system
        return result

    current = property(GetCurrent, SetCurrent)

    # Objects --------------------------------------------------------------------------------------

    def UpdateObjects(self) -> None:
        """
        Updates the units of all the registered objects. Also, remove dead objects from the list
        """
        current = self._current
        if current is not None:
            # remove dead references
            GetDefaultUnit = current.GetDefaultUnit
            for wrap in set(
                self._object_refs
            ):  # Note: iterate in a copy (just in case gc is triggered).
                obj = wrap.ref()
                if obj is not None:
                    # Update the object to match the current unit-system.
                    unit = GetDefaultUnit(obj.GetCategory())
                    if unit is not None:
                        obj.unit = unit

    def Register(self, obj: AbstractValueWithQuantityObject) -> None:
        """
        Keeps a weak reference to the object in an internal list.
        Whatever change in the unit system will be propagated for all objects in the list.

        :param obj:
            An object which represents a value associated with a unit.

        .. note:: This code should in general only be called from the constructor of an object to be
        tracked and only once (although adding an object more than once won't give any errors, it'll
        incur in more overhead because the object will be added more than once to the internal list).
        """
        self._object_refs.add(_IdentityWrap(obj, self))
        current = self._current
        if current is not None:
            # Update the object to match the current unit-system.
            unit = current.GetDefaultUnit(obj.GetCategory())
            if unit is not None:
                obj.unit = unit

    def GetUnitSystemById(self, id: str) -> IUnitSystem:
        """
        Returns the UnitSystem with the given id.

        :param id:
            The wished unit system id.

        :returns:
            The unit system with the given id.

        :raises ValueError:
            If no unit system with that id was found.
        """
        if id in self._unit_systems:
            return self._unit_systems[id]
        else:
            raise ValueError("No UnitSystem with id %r found" % id)

    def GetNewId(self) -> str:
        """
        Returns an unit system id that is not being used.

        :returns:
            The available id.
        """
        ids = {id for id in self._unit_systems}
        count = 1
        new_id = "%s %d" % ("system", count)
        while new_id in ids:
            count += 1
            new_id = "%s %d" % ("system", count)
        return new_id

    def ConvertToCurrent(
        self, category: str, unit: str, value: float, unit_database: Optional[UnitDatabase] = None
    ) -> Tuple[float, str]:
        """
        Converts the given value from the given unit to the default unit of the current unit system.
        If there is no current unit system, no conversion is performed.

        :param category:
            The unit's category.

        :param unit:
            The source unit.

        :param value:
            The value to convert.

        :param unit_database:
            The unit database to perform the conversion. If not given, use the singleton.

        :returns:
            The pair (value, unit), containing the converted value and target unit. Note that
            if there is no current unit system, the returned value and unit are the same as
            the input.
        """
        if unit_database is None:
            unit_database = UnitDatabase.GetSingleton()

        current = self.current
        if current is None or current.GetDefaultUnit(category) is None:
            return value, unit

        to_unit = current.GetDefaultUnit(category)
        converted_value = unit_database.Convert(category, unit, to_unit, value)
        return converted_value, to_unit

    def GetCategoryDefaultUnit(self, category: str) -> Optional[str]:
        """
        :param category:
            The category to obtain the default unit

        :returns:
            The default unit to display the given category or None if there is no default unit set.
        """
        current_system = self.GetCurrent()

        result = None
        if current_system is not None:
            default_unit = current_system.GetDefaultUnit(category)

            if default_unit is not None:
                result = default_unit

        return result

    def GetQuantityDefaultUnit(self, quantity: IQuantity) -> str:
        """
        Get the default unit to use with a given quantity. If there is any map explictly defined for
        the quantity category, then that map will be used. Otherwise we will use the quantity default
        unit.

        :param quantity:
            The quantity to obtain the default unit

        :return:
            The default unit to use with the given quantity
        """
        current_system = self.GetCurrent()

        # If no unit system was provided then let us use the quantity default unit
        result = quantity.GetUnit()

        if current_system is not None:
            default_unit = current_system.GetDefaultUnit(quantity.GetCategory())

            if default_unit is not None:
                result = default_unit

        return result

    def ConvertScalarToCurrent(
        self, scalar: "Scalar", unit_database: Optional[UnitDatabase] = None
    ) -> "Scalar":
        """
        Same as ConvertToCurrent but receives and returns a scalar instead of category, unit and value

        :param Scalar scalar:
            The scalar to be converted

        :param UnitDatabase unit_database:
            The unit database to perform the conversion. If not given, use the singleton.

        :rtype: Scalar
        :returns:
            A Scalar containing the converted value and target unit. Note that
            if there is no current unit system, the returned value and unit are the same as
            the input.
        """
        from barril.units import Scalar

        ret_tuple = self.ConvertToCurrent(
            scalar.GetCategory(), scalar.GetUnit(), scalar.GetValue(), unit_database
        )
        return Scalar(*ret_tuple)


class _IdentityWrap:
    """
    Helper class to remove an object from the unit system references.

    It's used so that we create a wrapper that'll give the __hash__ and __eq__ based on
    the object id.
    """

    __slots__ = ["unit_system", "ref"]

    def __init__(self, obj: AbstractValueWithQuantityObject, unit_system: UnitSystemManager):
        """
        :param obj:
            The object we'll be wrapping.

        :param unit_system:
            The unit system manager with the tracked object.
        """
        self.unit_system = unit_system
        self.ref = weakref.ref(obj, self._OnRefKilled)

    def _OnRefKilled(self, ref: object) -> None:
        """
        Called when a reference to an object with a unit is killed.

        :param ref:
            The weak-ref that was killed
        """
        self.unit_system._object_refs.remove(self)



"""
Locale-aware methods that convert to and from float values, trying to give the best representation
possible.
"""

PLUS_INFINITY = float("inf")
PLUS_INFINITY_STR = "+INF"

MINUS_INFINITY = float("-inf")
MINUS_INFINITY_STR = "-INF"

NAN = float("nan")
NAN_STR = "-1.#IND"


def FormatFloat(pattern: str, value: float, grouping: bool = False, use_locale: bool = True) -> str:
    """
    Formats the value given according to the current LC_NUMERIC setting. The format follows the
    conventions of the % operator. For floating point values, the decimal point is modified if
    appropriate. If grouping is True, the thousands separator set in locale settings is also taken
    into account.

    @note:
        This function is used to format float values. There are known issues like convert unit
        from scalar where value is "0.0", it can get representation "-0.0"

    :param pattern:
        The pattern used to format the value.

    :param value:
        The value to be formated.

    :param grouping:
        True if the thousands separator must be used on formating the number. The locale settings
        must be explicitly set in order to make sure the thousands separator will be applied. The
        default value for thousands separator is "", so that the value formated won't have thousands
        separator.

    :param use_locale:
        Use locale.format or str % operator (locale-independent output).

    :returns:
        The formatted value.
    """
    # Handling INFINITY
    # - locale.format tries to round the infinity value:
    #   ( ".3g", 9e999 ) ==> +1.#J
    import locale

    if value == PLUS_INFINITY:
        return PLUS_INFINITY_STR
    elif value == MINUS_INFINITY:
        return MINUS_INFINITY_STR
    elif value != value:  # Not a Number
        # numpy.isnan(x) is more readable, but it doesn't work here because FractionValue-s seem to
        # pass numpy arrays to FormatFloat, and numpy.isnan doesn't accept them.
        return NAN_STR

    # This removes the minus sign from the string representation of the float
    if value == 0.0:
        value = 0.0

    if use_locale:
        try:
            result = locale.format_string(pattern, value, grouping)
        except TypeError:
            # Python has a limitation to convert large float numbers to integer format.
            # To avoid this the values will be forced to int when requesting for the integer format.
            result = locale.format_string(pattern, int(value), grouping)
    else:
        result = pattern % value

    return result


def FloatFromString(str_value: str, use_locale: bool = True) -> float:
    """
    Converts the given string value into a float, taking in account the current locale.

    :param str_value:

    :returns:
        The equivalent float value

    :param use_locale:
        Use locale.atof or standard float conversion (default python output, locale-independent).

    :raises ValueError:
        If given string is not a valid float literal in the current locale
    """
    import locale

    if str_value.__class__ != str:
        from barril._util.types_ import CheckType

        CheckType(str_value, str)

    if str_value == PLUS_INFINITY_STR:
        return PLUS_INFINITY
    elif str_value == MINUS_INFINITY_STR:
        return MINUS_INFINITY
    elif str_value == NAN_STR:
        return NAN
    elif use_locale:
        return locale.atof(str_value)
    else:
        return float(str_value)



from typing import Any
from typing import cast
from typing import ClassVar
from typing import List
from typing import NoReturn
from typing import Optional
from typing import overload
from typing import Type
from typing import TypeVar
from typing import Union

from oop_ext.interface import ImplementsInterface

from ._quantity import ObtainQuantity
from ._quantity import Quantity
from .interfaces import IObjectWithQuantity
from .interfaces import IQuantity
from barril.units.unit_database import CategoryInfo
from barril.units.unit_database import UnitDatabase

__all__ = ["AbstractValueWithQuantityObject"]


T = TypeVar("T", bound="AbstractValueWithQuantityObject")


@ImplementsInterface(IObjectWithQuantity, IQuantity)
class AbstractValueWithQuantityObject:
    """
    This is an abstract class that provides a default implementation for having a class
    that has a value associated with a quantity.

    Subclasses should only provide a method for getting and setting the value, as it may be a
    numarray object, list of floats, etc (so, the user may want to call it GetValue, GetValues,
    GetNumarray, ...).

    .. see:: L{Scalar} for an implementation example
    """

    def __init__(
        self, category: Union[str, Quantity], value: Any = None, unit: Optional[str] = None
    ):
        # NOTE: we don't even try to add overloads here because each subclass defines their own
        # overloads, often in confusing ways; adding them here would just add to the
        # confusion.

        unit_database = UnitDatabase.GetSingleton()
        if isinstance(category, Quantity):
            quantity = category

            assert unit is None, "If quantity is given, the unit must not!"

            if value is None:
                value = self._GetDefaultValue(quantity.GetCategoryInfo())

        else:
            if not isinstance(category, str):
                # Support for creating a scalar as:
                # Scalar(10, 'm')
                # Scalar(10, 'm', 'length')
                value, unit, category = category, value, unit  # type:ignore[assignment]

            if value is None or unit is None:
                if value is None:
                    category_info = unit_database.GetCategoryInfo(category)
                    value = self._GetDefaultValue(category_info, unit)
                    if unit is None:
                        unit = category_info.default_unit
                        assert unit is not None
                else:
                    # Note: if the category is given with a value, the unit MUST also be given.
                    # (because if later we change the default unit, the original category would be wrong).
                    assert (
                        unit is not None
                    ), "If category and value are given, the unit must be specified too."

            quantity = ObtainQuantity(unit, category)  # type:ignore[arg-type, assignment]

        self._quantity = quantity
        self._InternalCreateWithQuantity(quantity, value, unit_database)

    def _GetDefaultValue(self, category_info: CategoryInfo, unit: Optional[str] = None) -> Any:
        raise NotImplementedError

    def _InternalCreateWithQuantity(
        self, quantity: Quantity, value: Any, unit_database: Optional[UnitDatabase] = None
    ) -> None:
        raise NotImplementedError

    def IsValid(self) -> bool:
        """
        :rtype: bool
        :returns:
            Checks if current state is valid.
            Should be implemented by each derived implementation.
        """
        if self.GetQuantity().IsDerived():
            return True
        try:
            self.CheckValidity()
        except ValueError:
            return False
        return True

    def CheckValidity(self) -> None:
        raise NotImplementedError

    def ConvertScalarValue(self, value: Any, to_unit: str) -> Any:
        raise NotImplementedError

    def HasCategory(self) -> bool:
        """
        Returns whether this instance has any associated category

        :returns:
            If there's some category composing this object
        """
        return len(self._quantity.GetComposingCategories()) > 0

    # UnitDatabase ---------------------------------------------------------------------------------
    def GetUnitDatabase(self) -> UnitDatabase:
        """
        :rtype: UnitDatabase
        :returns:
            The unit database that was used in the creation of this object.
        """
        return self._quantity.GetUnitDatabase()

    # Quantity -------------------------------------------------------------------------------------
    def GetQuantity(self) -> Quantity:
        """
        :rtype: Quantity
        :returns:
            The Quantity that is associated with this object.
        """
        return self._quantity

    # Category -------------------------------------------------------------------------------------
    def GetCategory(self) -> str:
        return self._quantity.GetCategory()

    category = property(GetCategory)

    # QuantityType ---------------------------------------------------------------------------------
    def GetQuantityType(self) -> str:
        return self._quantity.GetQuantityType()

    quantity_type = property(GetQuantityType)

    # Unit -----------------------------------------------------------------------------------------
    def GetUnit(self) -> str:
        return self._quantity.GetUnit()

    unit = property(GetUnit)

    # UnitDatabase Shortcuts -----------------------------------------------------------------------
    def GetUnitName(self) -> str:
        """
        :rtype: str
        :returns:
            Returns a user-friendly name for the given unit (i.e.: 'm' would me 'meters')
        """
        return self._quantity.GetUnitName()

    def GetValidUnits(self) -> List[str]:
        """
        :rtype: list(str)
        :returns:
            Returns a list with all the valid units for the category for this object + the current
            unit if it's not in the list of valid units (because it may be using a unit valid for
            the quantity type even if it doesn't match a unit in the category).
        """
        valid_units = self.GetUnitDatabase().GetValidUnits(self.GetCategory())
        current_unit = self.GetQuantity().GetUnit()
        if current_unit not in valid_units:
            valid_units.append(current_unit)
        return valid_units

    @classmethod
    def CreateWithQuantity(cls: Type[T], quantity: Quantity, *args: object, **kwargs: object) -> T:
        """
        This is a secondary interface for creating the object with an existing quantity.

        :param Quantity quantity:
            The quantity for this object

        @param args and kwargs: Those are dependent on the actual class -- this parameters
            are passed directly to _InternalCreateWithQuantity.
        """

        class Stub:
            pass

        stub = Stub()  # type:ignore[assignment]
        stub.__class__ = cls  # type:ignore[assignment]
        stub._InternalCreateWithQuantity(quantity, *args, **kwargs)  # type:ignore[attr-defined]
        return cast(T, stub)

    # Copy -----------------------------------------------------------------------------------------
    def Copy(self: T) -> T:
        """
        :returns:
            Returns self, it is immutable.
        """
        return self

    def CreateCopyInstance(self: T) -> T:
        """
        :rtype: Scalar
        :returns:
            Returns a new scalar that's a copy of this scalar.
        """
        return self

    def __copy__(self: T) -> T:
        """
        Copy protocol.
        """
        return self.Copy()

    def __deepcopy__(self: T, memo: object) -> T:
        """
        Copy protocol.
        """
        return self.Copy()

    def GetAbstractValue(self, unit: Optional[str] = None) -> Any:
        raise NotImplementedError

    def CreateCopy(
        self: T,
        value: Any = None,
        unit: Optional[str] = None,
        category: Optional[str] = None,
        **kwargs: object
    ) -> T:
        try:
            if value is None:
                value = self.GetAbstractValue(unit)

            if unit is None and category is None:
                return self.CreateWithQuantity(self._quantity, value=value, **kwargs)

            elif category is not None:
                if unit is None:
                    raise TypeError("If category is given, the unit must be specified too.")

                return self.CreateWithQuantity(
                    ObtainQuantity(unit, category), value=value, **kwargs
                )

            elif unit is not None:
                if self._quantity.GetCategory():
                    return self.CreateWithQuantity(
                        ObtainQuantity(unit, self._quantity.GetCategory()), value=value, **kwargs
                    )
                else:
                    # Handling empty quantity
                    return self.CreateWithQuantity(ObtainQuantity(unit), value=value, **kwargs)

            else:
                raise RuntimeError("Execution should never get here!")

        except TypeError as e:
            raise TypeError(
                "Error creating new instance of %s: %s\n"
                "(Should be overridden in '%s' class if it takes parameters in __init__)"
                % (self.__class__.__name__, e, self.__class__.__name__)
            )

    def __ne__(self, other: object) -> bool:
        return not self == other

    def __hash__(self) -> NoReturn:
        raise NotImplementedError(
            "Objects with a quantity are not hashable (as they're usually mutable)."
        )

    # Format ---------------------------------------------------------------------------------------

    FORMATTED_SUFFIX_FORMAT = " [%s]"

    @classmethod
    def GetFormattedSuffixFormat(cls) -> str:
        """
        Returns the formatted suffix for the unit.

        :returns:
            The formatted suffix
        """
        return cls.FORMATTED_SUFFIX_FORMAT

    @classmethod
    def SetFormattedSuffixFormat(cls, pattern: str) -> None:
        """
        Sets the format for the formatted text suffix, which may include the unit (Use "%s" to
        place the unit.).

        :param pattern:
            A format-like string containing one C{%s} format code
        """
        try:
            pattern % "unit"
        except TypeError as e:
            raise TypeError(
                "Incompatible pattern for Scalar suffix. Expected a format for a str value."
            ) from e
        cls.FORMATTED_SUFFIX_FORMAT = pattern

    def GetFormattedSuffix(self, unit: Optional[str] = None) -> str:
        """
        Returns the suffix for the formatted string using the current unit.

        :rtype: str
        :returns:
            The suffix
        """
        if unit is None:
            unit = self.GetUnit()
        return self.FORMATTED_SUFFIX_FORMAT % unit

    #: sentinel used to change the category of the object (see ChangeCategory in subclasses)
    DEFAULT_CATEGORY_VALUE: ClassVar[Any] = object()
    #: sentinel used to change the category of the object (see ChangeCategory in subclasses)
    CURRENT_VALUE: ClassVar[Any] = object()



from barril.units import Scalar


def testEmptyScalar() -> None:
    scalar = Scalar.CreateEmptyScalar()
    assert not scalar.HasCategory()
    assert scalar == scalar.Copy()

    scalar = scalar.CreateCopy(category="pressure", unit="psi")
    assert scalar.HasCategory()
    assert scalar.GetValue() == 0.0

    assert scalar == scalar.Copy()



from setuptools import find_packages
from setuptools import setup

with open("README.rst", encoding="UTF-8") as readme_file:
    readme = readme_file.read()

with open("CHANGELOG.rst", encoding="UTF-8") as changelog_file:
    history = changelog_file.read()

requirements = ["attrs>=18.1.0", "numpy>=1.11.0", "oop-ext>=1.1"]
extras_require = {
    "docs": ["sphinx >= 1.4", "sphinx_rtd_theme", "sphinx-autodoc-typehints", "typing_extensions"],
    "testing": [
        "codecov",
        "data-science-types",
        "mypy",
        "pre-commit",
        "pytest",
        "pytest-cov",
        "pytest-mock",
        "tox",
    ],
}

setup(
    author="ESSS",
    author_email="foss@esss.co",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    description="Python package to manage units for physical quantities",
    extras_require=extras_require,
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    python_requires=">=3.6",
    keywords="barril",
    name="barril",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    url="https://github.com/ESSS/barril",
    zip_safe=False,
)



"""
Extensions to python native types.
"""
from typing import Any
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union

_TRUE_VALUES = ["TRUE", "YES", "1"]
_FALSE_VALUES = ["FALSE", "NO", "0"]
_TRUE_FALSE_VALUES = _TRUE_VALUES + _FALSE_VALUES
_KNOWN_NUMBER_TYPES: Any = None


def _GetKnownNumberTypes() -> Tuple[type, ...]:
    """
    Dynamically obtain the tuple with the types considered number, including numpy.number if
    possible.

    This code replaces an old implementation with "code replacement". Not checked if we have any
    performance penalties.
    """
    result = {float, complex, int}
    try:
        import numpy
    except ImportError:
        pass
    else:
        result.add(numpy.number)
    return tuple(result)


def CheckType(
    object_: Any, type_: Union[type, Tuple[type, ...]], message: Optional[str] = None
) -> None:
    """
    Check if the given object is of the given type, raising a descriptive "TypeError" if it is
    not.

    :param object_:
        The object to check the type

    :param type_:
        The type or types to check.

    :param message:
        The error message shown if the check does not pass.
        By default, generates the following message:

            CheckType: Expecting <expected type(s)>, got <object type>: <object>

    @raise: TypeError:
        Raises type-error if the object does not matches the given type(s).
    """
    if not isinstance(object_, type_):
        # 001) The list of the types names.
        type_names_parts = [x if isinstance(x, str) else x.__name__ for x in MakeTuple(type_)]
        type_names = '" or "'.join(type_names_parts)

        # 002) Build the error message
        exception_message = 'CheckType: Expecting "{}", got "{}": {}'.format(
            type_names, object_.__class__.__name__, repr(object_)
        )

        # 003) Appends the user message at the end
        if message:
            exception_message += "\n%s" % message

        # 004) Raises the exception
        raise TypeError(exception_message)


def CheckFormatString(pattern: str, *arguments: Any) -> None:
    """
    Checks if the given format string (for instance, "%.g") is valid for the given arguments.
    :param pattern: a string in "%" format
    :param arguments: the arguments to check against
    :raises ValueError:
        if the format is invalid.
    """
    try:
        pattern % arguments
    except (TypeError, ValueError):
        raise ValueError("%r is not a valid format string." % pattern)


def IsNumber(v: object) -> bool:
    """
    Checks if the given value is a number

    :return:
        True if the given value is a number, False otherwise
    """
    global _KNOWN_NUMBER_TYPES
    if not _KNOWN_NUMBER_TYPES:
        _KNOWN_NUMBER_TYPES = _GetKnownNumberTypes()

    return isinstance(v, _KNOWN_NUMBER_TYPES)


def MakeTuple(object_: Any) -> Tuple[Any, ...]:
    """
    Returns the given object as a tuple, if it is not, creates one with it inside.

    @param:
        The object to tupleIZE
    """
    if isinstance(object_, tuple):
        return object_
    else:
        return (object_,)



from collections import OrderedDict
from typing import List
from typing import Tuple

import pytest
from pytest import approx

from barril import units
from barril.units import Array
from barril.units import InvalidUnitError
from barril.units import ObtainQuantity
from barril.units import Quantity
from barril.units import Scalar


def testEmptyArray() -> None:
    arr = Array.CreateEmptyArray()
    assert not arr.HasCategory()

    arr = arr.CreateCopy(category="temperature", unit="degC")
    assert arr.HasCategory()


def testValues() -> None:
    array = units.Array("length", values=[100, 150, 200], unit="m")
    assert array.unit == "m"
    assert array.values == [100, 150, 200]

    array = array.CreateCopy(values=list(range(100)))
    array = array.CreateCopy(unit="km")
    assert array.values == [x / 1000.0 for x in range(100)]
    assert 3.0 / 1000.0 == array[3]

    with pytest.raises(AttributeError):
        array.values = 10


def testArrayWithNaNs(unit_database_len) -> None:
    import numpy

    TEST_VALUES = [[numpy.nan, 0, 15, numpy.nan], [numpy.nan] * 5]
    for values in TEST_VALUES:
        units.Array(
            "flow rate", values=values, unit="m3/s"
        )  # check that this does not raise exceptions


def testStr() -> None:
    array = units.Array("length", values=[100, 150, 200], unit="m")
    assert str(array) == "100 150 200 [m]"


def testConvertValuesFromArray() -> None:
    def CheckConversion(values):
        _array = units.Array("length", values=values, unit="m")
        for converted, original in zip(_array.GetValues("km"), values):
            assert converted == original / 1000

    from numpy import array

    CheckConversion(array([0.0, 1.123, 2.456, 3.789]))
    CheckConversion(array([0.0, 0.0]))


def testValues2() -> None:
    array = units.Array("length", values=[100, 150, 200], unit="m")
    array2 = units.Array("length", values=(100, 150, 200), unit="m")
    assert array == array2


def testEquality() -> None:
    a1 = units.Array("length", values=list(range(100)), unit="m")

    assert a1 == units.Array("length", values=list(range(100)), unit="m")
    assert a1 != units.Array("temperature", values=list(range(100)), unit="degC")
    assert a1 != units.Array("length", values=[], unit="m")
    assert a1 != units.Array("length", values=list(range(100)), unit="km")


def testMultiDimensional() -> None:
    a1 = units.Array("length", values=[(1000, 1000), (300, 300)], unit="m")
    assert a1.GetValues("km") == [(1, 1), (0.3, 0.3)]

    assert (1000, 1000) == a1[0]
    assert (300, 300) == a1[1]

    a1 = a1.CreateCopy(unit="km")
    assert (1, 1) == a1[0]
    assert (0.3, 0.3) == a1[1]


def testCreateWithQuantity() -> None:
    q = ObtainQuantity("m", "length")
    a1 = units.Array(q, values=[(1000, 1000), (300, 300)])
    a2 = a1.CreateCopy(unit="km")

    assert a1.values == [(1000, 1000), (300, 300)]
    assert a2.values == [(1, 1), (0.3, 0.3)]


def testCategoryParameters(unit_database_len, mocker) -> None:
    unit_database = unit_database_len
    unit_database.AddUnit("temperature", "Kelvin Degrees", "degK", "%f - 273.5", "%f + 273.5")
    unit_database.AddCategory(
        "my temperature",
        "temperature",
        ["degC", "degK"],
        default_unit="degC",
        default_value=22.3,
        min_value=-10.0,
        max_value=213.4,
        is_min_exclusive=False,
        is_max_exclusive=True,
    )

    a1 = units.Array("my temperature", values=[0, 100], unit="degC")
    assert a1.IsValid()

    a2 = a1.CreateCopy(values=[0, 220])
    assert a2._is_valid is None
    assert a2._validity_exception is None

    mocker.patch.object(
        a2, "_DoValidateValues", autospec=True, side_effect=ValueError("my value error")
    )

    assert not a2.IsValid()
    assert a2._is_valid is False
    assert str(a2._validity_exception) == "my value error"
    assert a2._DoValidateValues.call_count == 1

    with pytest.raises(ValueError):
        a2.CheckValidity()

        # As result is already cached the "_DoValidateValues" will not be invoked.
        assert a2._DoValidateValues.call_count == 1

    a3 = units.Array("my temperature", values=[(-100, 50), (0, 2)], unit="degC")
    assert not a3.IsValid()


def testInvalidUnit() -> None:
    a1 = units.Array("length", values=[(1000, 1000), (300, 300)], unit="m")
    try:
        a1 = a1.CreateCopy(values=[(100, 200)], unit="foo")
    except InvalidUnitError:
        pass
    assert a1.values == [(1000, 1000), (300, 300)]
    assert a1.unit == "m"


def testGetValues() -> None:
    """
    Tests GetValues method return type when passing a 'unit'
    """
    array = units.Array("temperature", values=[0, 100], unit="degC")
    assert isinstance(array.GetValues(), list)
    assert isinstance(array.GetValues("degC"), list)

    array2 = units.Array("temperature", values=(0, 100), unit="degC")
    assert isinstance(array2.GetValues(), tuple)
    assert isinstance(array2.GetValues("degC"), tuple)
    assert array2[1] == 100

    array3 = units.Array[List[float]]("temperature", values=[], unit="degC")
    assert isinstance(array3.GetValues("degC"), list)


def testCopy() -> None:
    array = units.Array("length", values=[0, 100], unit="m")
    copy = array.Copy()
    assert copy is array

    # test copy of subclasses
    class MyArray(units.Array):
        pass

    array = MyArray("length", values=[0, 100], unit="m")
    copy = array.Copy()
    assert isinstance(copy, MyArray)


def testArrayOperations(unit_database_len_time) -> None:
    unit_database = unit_database_len_time

    m = Quantity.CreateDerived(OrderedDict([("Table size", ["m", 1])]))
    km_city = Quantity.CreateDerived(OrderedDict([("City size", ["km", 1])]))

    s1: Array[List[float]] = Array.CreateWithQuantity(m, [1])
    s2: Array[List[float]] = Array.CreateWithQuantity(km_city, [0.01])
    initial1 = s1.GetQuantity().GetComposingUnits()
    initial2 = s2.GetQuantity().GetComposingUnits()
    # Check that they doesn't raise ComposedUnitError
    s1.GetValues()
    s2.GetValues()

    quantity, value = unit_database.Multiply(m, km_city, 1, 0.01)
    assert initial1 == s1.GetQuantity().GetComposingUnits()
    assert initial2 == s2.GetQuantity().GetComposingUnits()
    calculated1: Array[List[float]] = Array.CreateWithQuantity(quantity, [value])

    array = s1 * s2
    assert str(array)
    assert calculated1 == s1 * s2

    quantity, value = unit_database.Sum(m, km_city, 1, 0.01)
    assert Array.CreateWithQuantity(quantity, [value]) == s1 + s2

    quantity, value = unit_database.Subtract(m, km_city, 1, 0.01)
    assert Array.CreateWithQuantity(quantity, [value]) == s1 - s2


def testDivision(unit_database_len_time) -> None:
    unit_database = unit_database_len_time
    m = Quantity.CreateDerived(OrderedDict([("Table size", ["m", 1])]))
    km_city = Quantity.CreateDerived(OrderedDict([("City size", ["km", 1])]))
    quantity, value = unit_database.Divide(m, km_city, 1, 0.01)
    calculated1: Array[List[float]] = Array.CreateWithQuantity(quantity, [value])
    s1: Array[List[float]] = Array.CreateWithQuantity(m, [1])
    s2: Array[List[float]] = Array.CreateWithQuantity(km_city, [0.01])
    assert calculated1 == s1 / s2


def testFloorDivision() -> None:
    a = Array([3.5, 4.2], "m")
    b = Array([100.0, 100.0], "cm")
    assert approx((a // b).GetValues()) == [3.0, 4.0]
    assert approx((350 // b).GetValues("1/cm")) == [3.0, 3.0]
    assert approx((a // 1.0).GetValues("m")) == [3.0, 4.0]


def testNumberOverArray() -> None:
    a = Array([2.0, 2.0], "m")
    b = Array([3.0, 3.0], "m")
    c = 1.0 / a

    assert approx(c.GetValues("1/m")) == [0.5, 0.5]
    assert approx((3.0 / a).GetValues("1/m")) == [1.5, 1.5]
    assert b / a == b * 1 / a == b * (1 / a)


def testNumberInteractions(unit_database_len_time) -> None:
    import numpy

    m = Quantity.CreateDerived(OrderedDict([("Table size", ["m", 1])]))
    s1: Array[List[int]] = Array.CreateWithQuantity(m, list(range(10)))
    s2: Array[List[int]] = Array.CreateWithQuantity(m, [x + x for x in range(10)])
    assert s1 == 0 + s1
    assert s1 == s1 + 0
    assert s2 == s1 + s1

    num_arr: Array[numpy.ndarray[numpy.int64]] = Array.CreateWithQuantity(
        m, numpy.array([x for x in range(10)])
    )
    sum_array = num_arr + s1
    assert isinstance(sum_array.values, numpy.ndarray)

    sum_array2: Array[numpy.ndarray[numpy.int64]] = num_arr + numpy.array([x for x in range(10)])
    assert isinstance(sum_array2.values, numpy.ndarray)

    tup_arr: Array[Tuple[int]] = Array.CreateWithQuantity(m, tuple([x for x in range(10)]))
    tup_arr = tup_arr + 1
    assert isinstance(tup_arr.values, tuple)
    assert tup_arr.values == (1, 2, 3, 4, 5, 6, 7, 8, 9, 10)


def testNumpyConversion() -> None:
    import numpy

    values = numpy.array([100, 200, 300])
    array = units.Array("length", values=values, unit="m")

    converted_values = array.GetValues("km")
    assert isinstance(converted_values, numpy.ndarray)
    assert list(numpy.array([0.1, 0.2, 0.3])) == list(converted_values)


def testNumpyMultiplicationOperation() -> None:
    import numpy

    units_array = units.Array("length", values=[100, 200, 300], unit="m")
    raw_array = [100, 200, 300]
    numpy_array = numpy.array([2.0, 0.5, 0.25])

    units_result = units_array * numpy_array
    raw_result = raw_array * numpy_array

    assert list(raw_result) == [200, 100, 75]
    assert list(units_result.GetValues("m")) == list(raw_result)


def testNoConversion() -> None:
    values = [100, 150, 200]
    array = units.Array("length", values=values, unit="m")
    assert array.GetValues() is values


def testNoConversionWhenUsingSameUnit() -> None:
    import numpy

    values = numpy.array([100, 150, 200])
    array = units.Array(values, "m")
    assert array.GetValues("m") is values


def testZeroDimensionalNumpyArray() -> None:
    import numpy

    values = numpy.array(1000.0)  # type:ignore[call-overload]
    array = units.Array(values, "m")
    assert array.GetValues("m") == numpy.array(1000.0)  # type:ignore[call-overload]
    assert array.GetValues("km") == numpy.array(1.0)  # type:ignore[call-overload]


def testReadOnlyQuantity() -> None:
    quantity = ObtainQuantity("m", "length")
    array = units.Array(quantity, values=[1, 2, 3])

    array = array.CreateCopy(unit="km")
    assert array.unit == "km"
    assert approx(array.GetValues()) == [0.001, 0.002, 0.003]


def testCopyPropertiesAndValidation(unit_database_len) -> None:
    # Not raises exception because by default validation is False on the copy operation
    array_source = Array("flow rate", values=[-1, -2, -3], unit="m3/s")
    assert not array_source.IsValid()

    array_dest = array_source.CreateCopy()
    assert array_dest.values == [-1, -2, -3]


def testDefaultValues(unit_database_len) -> None:
    # Not raises exception because by default validation is False on the copy operation
    array = Array[List[float]]("flow rate")
    assert array.values == []

    array = Array(ObtainQuantity("m"))
    assert array.values == []

    with pytest.raises(AssertionError):
        Array(ObtainQuantity("m"), unit="m")  # type:ignore[call-overload]


class TestFromScalar:
    def test_create_array_informing_category(self) -> None:
        array_molar_fraction = Array.FromScalars(
            scalars=[Scalar(1, "-"), Scalar(2, "-")], category="percentage"
        )
        assert array_molar_fraction.values == [1, 2]
        assert array_molar_fraction.unit == "-"
        assert array_molar_fraction.category == "percentage"

    def test_create_array_informing_unit(self) -> None:
        array_in_cm = Array.FromScalars(scalars=[Scalar(1, "m"), Scalar(2, "m")], unit="cm")
        assert array_in_cm.values == [100.0, 200.0]
        assert array_in_cm.unit == "cm"
        assert array_in_cm.category == "length"

    def test_create_array_from_list_of_scalar(self) -> None:
        array_in_m = Array.FromScalars(scalars=iter([Scalar(1, "m"), Scalar(2, "m")]))
        assert array_in_m.values == [1, 2]
        assert array_in_m.unit == "m"
        assert array_in_m.category == "length"

    def test_check_empty_array(self) -> None:
        assert Array.FromScalars(scalars=[]) == Array.CreateEmptyArray()
        assert Array.FromScalars(scalars=[], unit="m") == Array([], "m")

        expected_msg = "If category and value are given, the unit must be specified too."
        with pytest.raises(AssertionError, match=expected_msg):
            Array.FromScalars(scalars=[], category="length")

    def test_check_array_with_different_units(self) -> None:
        with pytest.raises(InvalidUnitError):
            Array.FromScalars(scalars=[Scalar(1, "m"), Scalar(1, "kg")])
